"""Classes to manage connection establishment under RFC 23 (DID exchange)."""

import json
import logging
from typing import List, Optional, Sequence, Tuple, Union

from did_peer_4 import LONG_PATTERN, long_to_short

from ....admin.server import AdminResponder
from ....connections.base_manager import BaseConnectionManager, BaseConnectionManagerError
from ....connections.models.conn_record import ConnRecord
from ....connections.models.connection_target import ConnectionTarget
from ....core.error import BaseError
from ....core.oob_processor import OobMessageProcessor
from ....core.profile import Profile
from ....did.did_key import DIDKey
from ....messaging.decorators.attach_decorator import AttachDecorator
from ....messaging.responder import BaseResponder
from ....storage.error import StorageNotFoundError
from ....transport.inbound.receipt import MessageReceipt
from ....wallet.base import BaseWallet
from ....wallet.did_method import SOV
from ....wallet.did_posture import DIDPosture
from ....wallet.error import WalletError
from ....wallet.key_type import ED25519
from ...coordinate_mediation.v1_0.manager import MediationManager
from ...coordinate_mediation.v1_0.models.mediation_record import MediationRecord
from ...discovery.v2_0.manager import V20DiscoveryMgr
from ...out_of_band.v1_0.messages.invitation import (
    InvitationMessage as OOBInvitationMessage,
)
from ...out_of_band.v1_0.messages.service import Service as OOBService
from .message_types import ARIES_PROTOCOL as DIDEX_1_1
from .message_types import DIDEX_1_0
from .messages.complete import DIDXComplete
from .messages.problem_report import DIDXProblemReport, ProblemReportReason
from .messages.request import DIDXRequest
from .messages.response import DIDXResponse


class DIDXManagerError(BaseError):
    """Connection error."""


class LegacyHandlingFallback(DIDXManagerError):
    """Raised when a request cannot be completed using updated semantics.

    Triggers falling back to legacy handling.
    """


class DIDXManager(BaseConnectionManager):
    """Class for managing connections under RFC 23 (DID exchange)."""

    SUPPORTED_USE_DID_METHODS = ("did:peer:2", "did:peer:4")

    def __init__(self, profile: Profile):
        """Initialize a DIDXManager.

        Args:
            profile: The profile for this did exchange manager
        """
        self._profile = profile
        self._logger = logging.getLogger(__name__)
        super().__init__(self._profile)

    @property
    def profile(self) -> Profile:
        """Accessor for the current profile.

        Returns:
            The profile for this did exchange manager

        """
        return self._profile

    async def receive_invitation(
        self,
        invitation: OOBInvitationMessage,
        their_public_did: Optional[str] = None,
        auto_accept: Optional[bool] = None,
        alias: Optional[str] = None,
        mediation_id: Optional[str] = None,
        protocol: Optional[str] = None,
    ) -> ConnRecord:  # leave in didexchange as it uses a responder: not out-of-band
        """Create a new connection record to track a received invitation.

        Args:
            invitation: invitation to store
            their_public_did: their public DID
            auto_accept: set to auto-accept invitation (None to use config)
            alias: optional alias to set on record
            mediation_id: record id for mediation with routing_keys, service endpoint
            protocol: The protocol to use for the connection

        Returns:
            The new `ConnRecord` instance

        """
        if not invitation.services:
            raise DIDXManagerError(
                "Invitation must contain service blocks or service DIDs"
            )
        else:
            for s in invitation.services:
                if isinstance(s, OOBService):
                    if not s.recipient_keys or not s.service_endpoint:
                        raise DIDXManagerError(
                            "All service blocks in invitation with no service DIDs "
                            "must contain recipient key(s) and service endpoint(s)"
                        )

        accept = (
            ConnRecord.ACCEPT_AUTO
            if (
                auto_accept
                or (
                    auto_accept is None
                    and self.profile.settings.get("debug.auto_accept_invites")
                )
            )
            else ConnRecord.ACCEPT_MANUAL
        )
        protocol = protocol or DIDEX_1_0
        if protocol not in ConnRecord.SUPPORTED_PROTOCOLS:
            raise DIDXManagerError(f"Unexpected protocol: {protocol}")

        service_item = invitation.services[0]
        # Create connection record
        conn_rec = ConnRecord(
            invitation_key=(
                DIDKey.from_did(service_item.recipient_keys[0]).public_key_b58
                if isinstance(service_item, OOBService)
                else None
            ),
            invitation_msg_id=invitation._id,
            their_label=invitation.label,
            their_role=ConnRecord.Role.RESPONDER.rfc23,
            state=ConnRecord.State.INVITATION.rfc160,
            accept=accept,
            alias=alias,
            their_public_did=their_public_did,
            connection_protocol=protocol,
        )

        async with self.profile.session() as session:
            await conn_rec.save(
                session,
                reason="Created new connection record from invitation",
                log_params={
                    "invitation": invitation,
                    "their_role": ConnRecord.Role.RESPONDER.rfc23,
                },
            )

            # Save the invitation for later processing
            await conn_rec.attach_invitation(session, invitation)
            if not conn_rec.invitation_key and conn_rec.their_public_did:
                targets = await self.resolve_connection_targets(conn_rec.their_public_did)
                conn_rec.invitation_key = targets[0].recipient_keys[0]

        await self._route_manager.save_mediator_for_connection(
            self.profile, conn_rec, mediation_id=mediation_id
        )

        if conn_rec.accept == ConnRecord.ACCEPT_AUTO:
            request = await self.create_request(conn_rec, mediation_id=mediation_id)
            base_responder = self.profile.inject(BaseResponder)
            responder = AdminResponder(self.profile, base_responder.send_fn)
            if responder:
                await responder.send_reply(
                    request,
                    connection_id=conn_rec.connection_id,
                )

                conn_rec.state = ConnRecord.State.REQUEST.rfc160
                async with self.profile.session() as session:
                    await conn_rec.save(session, reason="Sent connection request")
        else:
            self._logger.debug("Connection invitation will await acceptance")

        return conn_rec

    async def create_request_implicit(
        self,
        their_public_did: str,
        my_label: Optional[str] = None,
        my_endpoint: Optional[str] = None,
        mediation_id: Optional[str] = None,
        use_public_did: bool = False,
        alias: Optional[str] = None,
        goal_code: Optional[str] = None,
        goal: Optional[str] = None,
        auto_accept: bool = False,
        protocol: Optional[str] = None,
        use_did: Optional[str] = None,
        use_did_method: Optional[str] = None,
    ) -> ConnRecord:
        """Create and send a request against a public DID only (no explicit invitation).

        Args:
            their_public_did (str): Public DID to which to request a connection.
            my_label (Optional[str], optional): My label for the request.
                Defaults to None.
            my_endpoint (Optional[str], optional): My endpoint. Defaults to None.
            mediation_id (Optional[str], optional): Record ID for mediation with routing
                keys, service endpoint. Defaults to None.
            use_public_did (bool, optional): Use my public DID for this connection.
                Defaults to False.
            alias (Optional[str], optional): Alias for the connection. Defaults to None.
            goal_code (Optional[str], optional): Optional self-attested code for sharing
                intent of connection. Defaults to None.
            goal (Optional[str], optional): Optional self-attested string for sharing
                intent of connection. Defaults to None.
            auto_accept (bool, optional): Auto-accept a corresponding connection request.
                Defaults to False.
            protocol (Optional[str], optional): Protocol to use for the connection.
                Defaults to None.
            use_did (Optional[str], optional): Local DID to use for the connection.
                Defaults to None.
            use_did_method (Optional[str], optional): DID method to use for the
                connection. Defaults to None.

        Returns:
            ConnRecord: The new `ConnRecord` instance representing the connection.

        Raises:
            DIDXManagerError: If both `use_did` and `use_did_method` are specified, or
                if both `use_public_did` and `use_did` are specified, or if both
                `use_public_did` and `use_did_method` are specified.
            WalletError: If no public DID is configured.
            DIDXManagerError: If a connection already exists for the specified
                `their_public_did` and `my_did`.

        """

        if use_did and use_did_method:
            raise DIDXManagerError("Cannot specify both use_did and use_did_method")

        if use_public_did and use_did:
            raise DIDXManagerError("Cannot specify both use_public_did and use_did")

        if use_public_did and use_did_method:
            raise DIDXManagerError(
                "Cannot specify both use_public_did and use_did_method"
            )

        my_info = None
        async with self.profile.session() as session:
            wallet = session.inject(BaseWallet)
            if use_public_did:
                my_info = await wallet.get_public_did()
                if not my_info:
                    raise WalletError("No public DID configured")
                if (
                    my_info.did == their_public_did
                    or f"did:sov:{my_info.did}" == their_public_did
                ):
                    raise DIDXManagerError(
                        "Cannot connect to yourself through public DID"
                    )
            elif use_did:
                my_info = await wallet.get_local_did(use_did)

            if my_info:
                try:
                    await ConnRecord.retrieve_by_did(
                        session,
                        their_did=their_public_did,
                        my_did=my_info.did,
                    )
                    raise DIDXManagerError(
                        "Connection already exists for their_did "
                        f"{their_public_did} and my_did {my_info.did}"
                    )
                except StorageNotFoundError:
                    pass

        auto_accept = bool(
            auto_accept
            or (
                auto_accept is None
                and self.profile.settings.get("debug.auto_accept_requests")
            )
        )
        protocol = protocol or DIDEX_1_0
        conn_rec = ConnRecord(
            my_did=(
                my_info.did if my_info else None
            ),  # create-request will fill in on local DID creation
            their_did=their_public_did,
            their_label=None,
            their_role=ConnRecord.Role.RESPONDER.rfc23,
            invitation_key=None,
            invitation_msg_id=None,
            alias=alias,
            their_public_did=their_public_did,
            connection_protocol=protocol,
            accept=ConnRecord.ACCEPT_AUTO if auto_accept else ConnRecord.ACCEPT_MANUAL,
        )
        request = await self.create_request(  # saves and updates conn_rec
            conn_rec=conn_rec,
            my_label=my_label,
            my_endpoint=my_endpoint,
            mediation_id=mediation_id,
            goal_code=goal_code,
            goal=goal,
            use_did_method=use_did_method,
        )
        conn_rec.request_id = request._id
        conn_rec.state = ConnRecord.State.REQUEST.rfc160
        async with self.profile.session() as session:
            await conn_rec.save(session, reason="Created connection request")
        responder = self.profile.inject_or(BaseResponder)
        if responder:
            await responder.send(request, connection_id=conn_rec.connection_id)

        return conn_rec

    async def create_request(
        self,
        conn_rec: ConnRecord,
        my_label: Optional[str] = None,
        my_endpoint: Optional[str] = None,
        mediation_id: Optional[str] = None,
        goal_code: Optional[str] = None,
        goal: Optional[str] = None,
        use_did_method: Optional[str] = None,
    ) -> DIDXRequest:
        """Create a new connection request for a previously-received invitation.

        Args:
            conn_rec: The `ConnRecord` representing the invitation to accept
            my_label: My label for request
            my_endpoint: My endpoint
            mediation_id: The record id for mediation that contains routing_keys and
                service endpoint
            goal_code: Optional self-attested code for sharing intent of connection
            goal: Optional self-attested string for sharing intent of connection
            use_did_method: The DID method to use for the connection
        Returns:
            A new `DIDXRequest` message to send to the other agent

        """
        if use_did_method and use_did_method not in self.SUPPORTED_USE_DID_METHODS:
            raise DIDXManagerError(
                f"Unsupported use_did_method: {use_did_method}. Supported methods: "
                f"{self.SUPPORTED_USE_DID_METHODS}"
            )

        # Mediation Support
        mediation_records = await self._route_manager.mediation_records_for_connection(
            self.profile,
            conn_rec,
            mediation_id,
            or_default=True,
        )

        if my_endpoint:
            my_endpoints = [my_endpoint]
        else:
            my_endpoints = []
            default_endpoint = self.profile.settings.get("default_endpoint")
            if default_endpoint:
                my_endpoints.append(default_endpoint)
            my_endpoints.extend(self.profile.settings.get("additional_endpoints", []))

        if not my_label:
            my_label = self.profile.settings.get("default_label")
            assert my_label

        did_url = None
        if conn_rec.their_public_did is not None:
            try:
                services = await self.resolve_didcomm_services(conn_rec.their_public_did)
            except BaseConnectionManagerError as e:
                raise DIDXManagerError(
                    "Failed to resolve DIDComm services from "
                    f"{conn_rec.their_public_did}: {e}"
                ) from e
            if services:
                did_url = services[0].id

        pthid = conn_rec.invitation_msg_id or did_url

        if conn_rec.connection_protocol == DIDEX_1_0:
            did, attach = await self._legacy_did_with_attached_doc(
                conn_rec, my_endpoints, mediation_records
            )
        else:
            if conn_rec.accept == ConnRecord.ACCEPT_AUTO or use_did_method is None:
                # If we're auto accepting or engaging in 1.1 without setting a
                # use_did_method, default to did:peer:4
                use_did_method = "did:peer:4"
            try:
                did, attach = await self._qualified_did_with_fallback(
                    conn_rec,
                    my_endpoints,
                    mediation_records,
                    use_did_method,
                )
            except LegacyHandlingFallback:
                did, attach = await self._legacy_did_with_attached_doc(
                    conn_rec, my_endpoints, mediation_records
                )

        request = DIDXRequest(
            label=my_label,
            did=did,
            did_doc_attach=attach,
            goal=goal,
            goal_code=goal_code,
        )

        if conn_rec.connection_protocol == DIDEX_1_0:
            request.assign_version("1.0")

        request.assign_thread_id(thid=request._id, pthid=pthid)

        # Update connection state
        conn_rec.request_id = request._id
        conn_rec.state = ConnRecord.State.REQUEST.rfc160
        async with self.profile.session() as session:
            await conn_rec.save(session, reason="Created connection request")

        # Idempotent; if routing has already been set up, no action taken
        await self._route_manager.route_connection_as_invitee(
            self.profile, conn_rec, mediation_records
        )

        return request

    async def _qualified_did_with_fallback(
        self,
        conn_rec: ConnRecord,
        my_endpoints: Sequence[str],
        mediation_records: List[MediationRecord],
        use_did_method: Optional[str] = None,
        signing_key: Optional[str] = None,
    ) -> Tuple[str, Optional[AttachDecorator]]:
        """Create DID Exchange request using a qualified DID.

        Fall back to unqualified DID if settings don't cause did:peer emission.
        """
        if conn_rec.my_did:  # DID should be public or qualified
            async with self.profile.session() as session:
                wallet = session.inject(BaseWallet)
                my_info = await wallet.get_local_did(conn_rec.my_did)

            posture = DIDPosture.get(my_info.metadata)
            if posture not in (
                DIDPosture.PUBLIC,
                DIDPosture.POSTED,
            ) and not my_info.did.startswith("did:"):
                raise LegacyHandlingFallback(
                    "DID has been previously set and not public or qualified"
                )
        elif use_did_method == "did:peer:4":
            my_info = await self.create_did_peer_4(my_endpoints, mediation_records)
            conn_rec.my_did = my_info.did
        elif use_did_method == "did:peer:2":
            my_info = await self.create_did_peer_2(my_endpoints, mediation_records)
            conn_rec.my_did = my_info.did
        else:
            # We shouldn't hit this condition in practice
            raise LegacyHandlingFallback(
                "Use of qualified DIDs not set according to settings"
            )

        did = conn_rec.my_did
        assert did, "DID must be set on connection record"
        if not did.startswith("did:"):
            did = f"did:sov:{did}"

        attach = None
        if signing_key:
            attach = AttachDecorator.data_base64_string(did)
            async with self.profile.session() as session:
                wallet = session.inject(BaseWallet)
                await attach.data.sign(signing_key, wallet)

        return did, attach

    async def _legacy_did_with_attached_doc(
        self,
        conn_rec: ConnRecord,
        my_endpoints: Sequence[str],
        mediation_records: List[MediationRecord],
        invitation_key: Optional[str] = None,
    ) -> Tuple[str, Optional[AttachDecorator]]:
        """Create a DID Exchange request using an unqualified DID."""
        if conn_rec.my_did:
            async with self.profile.session() as session:
                wallet = session.inject(BaseWallet)
                my_info = await wallet.get_local_did(conn_rec.my_did)
        else:
            async with self.profile.session() as session:
                wallet = session.inject(BaseWallet)
                my_info = await wallet.create_local_did(
                    method=SOV,
                    key_type=ED25519,
                )
                conn_rec.my_did = my_info.did

        posture = DIDPosture.get(my_info.metadata)
        if posture in (
            DIDPosture.PUBLIC,
            DIDPosture.POSTED,
        ):
            return my_info.did, None

        did_doc = await self.create_did_document(
            my_info,
            my_endpoints,
            mediation_records=mediation_records,
        )
        attach = AttachDecorator.data_base64(did_doc.serialize())

        async with self.profile.session() as session:
            wallet = session.inject(BaseWallet)
            await attach.data.sign(invitation_key or my_info.verkey, wallet)

        return my_info.did, attach

    async def receive_request(
        self,
        request: DIDXRequest,
        recipient_did: str,
        recipient_verkey: Optional[str] = None,
        alias: Optional[str] = None,
        auto_accept_implicit: Optional[bool] = None,
    ) -> ConnRecord:
        """Receive and store a connection request.

        Args:
            request: The `DIDXRequest` to accept
            recipient_did: The (unqualified) recipient DID
            recipient_verkey: The recipient verkey: None for public recipient DID
            my_endpoint: My endpoint
            alias: Alias for the connection
            auto_accept_implicit: Auto-accept request against implicit invitation
        Returns:
            The new or updated `ConnRecord` instance

        """
        ConnRecord.log_state(
            "Receiving connection request",
            {"request": request},
            settings=self.profile.settings,
        )

        if recipient_verkey:
            conn_rec = await self._receive_request_pairwise_did(request, alias)
        else:
            conn_rec = await self._receive_request_public_did(
                request, recipient_did, alias, auto_accept_implicit
            )

        # Clean associated oob record if not needed anymore
        oob_processor = self.profile.inject(OobMessageProcessor)
        await oob_processor.clean_finished_oob_record(self.profile, request)

        return conn_rec

    async def _receive_request_pairwise_did(
        self,
        request: DIDXRequest,
        alias: Optional[str] = None,
    ) -> ConnRecord:
        """Receive a DID Exchange request against a pairwise (not public) DID."""
        if not request._thread.pthid:
            raise DIDXManagerError("DID Exchange request missing parent thread ID")

        async with self.profile.session() as session:
            conn_rec = await ConnRecord.retrieve_by_invitation_msg_id(
                session=session,
                invitation_msg_id=request._thread.pthid,
                their_role=ConnRecord.Role.REQUESTER.rfc23,
            )

        if not conn_rec:
            raise DIDXManagerError(
                "Pairwise requests must be against explicit invitations that have not "
                "been previously consumed"
            )

        if conn_rec.is_multiuse_invitation:
            conn_rec = await self._derive_new_conn_from_multiuse_invitation(conn_rec)

        conn_rec.their_label = request.label
        if alias:
            conn_rec.alias = alias
        conn_rec.their_did = request.did
        conn_rec.state = ConnRecord.State.REQUEST.rfc160
        conn_rec.request_id = request._id
        conn_rec.connection_protocol = self._handshake_protocol_to_use(request)

        # TODO move to common method or add to transaction?
        await self._extract_and_record_did_doc_info(request)

        async with self.profile.transaction() as txn:
            # Attach the connection request so it can be found and responded to
            await conn_rec.save(txn, reason="Received connection request from invitation")
            await conn_rec.attach_request(txn, request)
            await txn.commit()

        return conn_rec

    def _handshake_protocol_to_use(self, request: DIDXRequest):
        """Determine the connection protocol to use based on the request.

        If we support it, we'll send it. If we don't, we'll try didexchange/1.1.
        """
        protocol = f"{request._type.protocol}/{request._type.version}"
        if protocol in ConnRecord.SUPPORTED_PROTOCOLS:
            return protocol

        return DIDEX_1_1

    async def _receive_request_public_did(
        self,
        request: DIDXRequest,
        recipient_did: str,
        alias: Optional[str] = None,
        auto_accept_implicit: Optional[bool] = None,
    ) -> ConnRecord:
        """Receive a DID Exchange request against a public DID."""
        if not self.profile.settings.get("public_invites"):
            raise DIDXManagerError(
                "Public invitations are not enabled: connection request refused"
            )

        async with self.profile.session() as session:
            wallet = session.inject(BaseWallet)
            public_did_info = await wallet.get_local_did(recipient_did)

        if DIDPosture.get(public_did_info.metadata) not in (
            DIDPosture.PUBLIC,
            DIDPosture.POSTED,
        ):
            raise DIDXManagerError(f"Request DID {recipient_did} is not public")

        if request._thread.pthid:
            # Invitation was explicit
            async with self.profile.session() as session:
                conn_rec = await ConnRecord.retrieve_by_invitation_msg_id(
                    session=session,
                    invitation_msg_id=request._thread.pthid,
                    their_role=ConnRecord.Role.REQUESTER.rfc23,
                )
        else:
            # Invitation was implicit
            conn_rec = None

        if conn_rec and conn_rec.is_multiuse_invitation:
            conn_rec = await self._derive_new_conn_from_multiuse_invitation(conn_rec)

        save_reason = None
        if conn_rec:
            conn_rec.their_label = request.label
            if alias:
                conn_rec.alias = alias
            conn_rec.their_did = request.did
            conn_rec.state = ConnRecord.State.REQUEST.rfc160
            conn_rec.request_id = request._id
            save_reason = "Received connection request from invitation to public DID"
        else:
            # request is against implicit invitation on public DID
            if not self.profile.settings.get("requests_through_public_did"):
                raise DIDXManagerError(
                    "Unsolicited connection requests to public DID is not enabled"
                )

            auto_accept = bool(
                auto_accept_implicit
                or (
                    auto_accept_implicit is None
                    and self.profile.settings.get("debug.auto_accept_requests", False)
                )
            )

            conn_rec = ConnRecord(
                my_did=None,  # Defer DID creation until create_response
                accept=(
                    ConnRecord.ACCEPT_AUTO if auto_accept else ConnRecord.ACCEPT_MANUAL
                ),
                their_did=request.did,
                their_label=request.label,
                alias=alias,
                their_role=ConnRecord.Role.REQUESTER.rfc23,
                invitation_key=public_did_info.verkey,
                invitation_msg_id=None,
                request_id=request._id,
                state=ConnRecord.State.REQUEST.rfc160,
            )
            save_reason = "Received connection request from public DID"

        conn_rec.connection_protocol = self._handshake_protocol_to_use(request)

        # TODO move to common method or add to transaction?
        await self._extract_and_record_did_doc_info(request)

        async with self.profile.transaction() as txn:
            # Attach the connection request so it can be found and responded to
            await conn_rec.save(txn, reason=save_reason)
            await conn_rec.attach_request(txn, request)
            await txn.commit()

        return conn_rec

    async def _extract_and_record_did_doc_info(self, request: DIDXRequest):
        """Extract and record DID Document information from the DID Exchange request.

        Extracting this info enables us to correlate messages from these keys back to a
        connection when we later receive inbound messages.
        """
        if request.did_doc_attach and request.did_doc_attach.data:
            self._logger.debug("Received DID Doc attachment in request")
            async with self.profile.session() as session:
                wallet = session.inject(BaseWallet)
                conn_did_doc = await self.verify_diddoc(wallet, request.did_doc_attach)
                await self.store_did_document(conn_did_doc)

            # Special case: legacy DIDs were unqualified in request, qualified in doc
            if request.did and not request.did.startswith("did:"):
                did_to_check = f"did:sov:{request.did}"
            else:
                did_to_check = request.did

            if did_to_check != conn_did_doc["id"]:
                raise DIDXManagerError(
                    (
                        f"Connection DID {request.did} does not match "
                        f"DID Doc id {conn_did_doc['id']}"
                    ),
                    error_code=ProblemReportReason.REQUEST_NOT_ACCEPTED.value,
                )
        else:
            if request.did is None:
                raise DIDXManagerError("No DID in request")

            self._logger.debug(
                "No DID Doc attachment in request; doc will be resolved from DID"
            )
            await self.record_keys_for_resolvable_did(request.did)

    async def _derive_new_conn_from_multiuse_invitation(
        self, conn_rec: ConnRecord
    ) -> ConnRecord:
        """Derive a new connection record from a multi-use invitation.

        Multi-use invitations are tracked using a connection record. When a connection
        is formed through a multi-use invitation conn rec, a new record for the resulting
        connection is required. The original multi-use invitation record is retained
        until deleted by the user.
        """
        new_conn_rec = ConnRecord(
            invitation_key=conn_rec.invitation_key,
            invitation_msg_id=conn_rec.invitation_msg_id,
            state=ConnRecord.State.INIT.rfc160,
            accept=conn_rec.accept,
            their_role=conn_rec.their_role,
        )
        async with self.profile.session() as session:
            # TODO: Suppress the event that gets emitted here?
            await new_conn_rec.save(
                session,
                reason="Created new connection record from multi-use invitation",
            )

        # Transfer metadata from multi-use to new connection
        # Must come after save so there's an ID to associate with metadata
        async with self.profile.session() as session:
            for key, value in (await conn_rec.metadata_get_all(session)).items():
                await new_conn_rec.metadata_set(session, key, value)

        return new_conn_rec

    async def create_response(
        self,
        conn_rec: ConnRecord,
        my_endpoint: Optional[str] = None,
        mediation_id: Optional[str] = None,
        use_public_did: Optional[bool] = None,
    ) -> DIDXResponse:
        """Create a connection response for a received connection request.

        Args:
            conn_rec: The `ConnRecord` with a pending connection request
            my_endpoint: Current agent endpoint
            mediation_id: The record id for mediation that contains routing_keys and
                service endpoint
            use_public_did: Use public DID for connection

        Returns:
            New `DIDXResponse` message

        """
        ConnRecord.log_state(
            "Creating connection response",
            {"connection_id": conn_rec.connection_id},
            settings=self.profile.settings,
        )

        mediation_records = await self._route_manager.mediation_records_for_connection(
            self.profile, conn_rec, mediation_id
        )

        if ConnRecord.State.get(conn_rec.state) is not ConnRecord.State.REQUEST:
            raise DIDXManagerError(
                f"Connection not in state {ConnRecord.State.REQUEST.rfc23}"
            )
        async with self.profile.session() as session:
            request = await conn_rec.retrieve_request(session)

        if my_endpoint:
            my_endpoints = [my_endpoint]
        else:
            my_endpoints = []
            default_endpoint = self.profile.settings.get("default_endpoint")
            if default_endpoint:
                my_endpoints.append(default_endpoint)
            my_endpoints.extend(self.profile.settings.get("additional_endpoints", []))

        if conn_rec.their_did and conn_rec.their_did.startswith("did:peer:2"):
            use_did_method = "did:peer:2"
        elif conn_rec.their_did and conn_rec.their_did.startswith("did:peer:4"):
            use_did_method = "did:peer:4"
        elif conn_rec.their_did and conn_rec.their_did.startswith("did:peer:1"):
            use_did_method = "did:peer:4"
        else:
            use_did_method = None

        if use_public_did:
            async with self.profile.session() as session:
                wallet = session.inject(BaseWallet)
                public_info = await wallet.get_public_did()
            if public_info:
                conn_rec.my_did = public_info.did
            else:
                raise DIDXManagerError("No public DID configured")

        if conn_rec.connection_protocol == DIDEX_1_0:
            did, attach = await self._legacy_did_with_attached_doc(
                conn_rec,
                my_endpoints,
                mediation_records,
                invitation_key=conn_rec.invitation_key,
            )
            response = DIDXResponse(did=did, did_doc_attach=attach)
            response.assign_version("1.0")
        else:
            try:
                did, attach = await self._qualified_did_with_fallback(
                    conn_rec,
                    my_endpoints,
                    mediation_records,
                    use_did_method=use_did_method,
                    signing_key=conn_rec.invitation_key,
                )
                response = DIDXResponse(did=did, did_rotate_attach=attach)
            except LegacyHandlingFallback:
                did, attach = await self._legacy_did_with_attached_doc(
                    conn_rec, my_endpoints, mediation_records, conn_rec.invitation_key
                )
                response = DIDXResponse(did=did, did_doc_attach=attach)

        # Idempotent; if routing has already been set up, no action taken
        await self._route_manager.route_connection_as_inviter(
            self.profile, conn_rec, mediation_records
        )

        # Assign thread information
        response.assign_thread_from(request)
        response.assign_trace_from(request)

        # Update connection state
        conn_rec.state = ConnRecord.State.RESPONSE.rfc23
        async with self.profile.session() as session:
            await conn_rec.save(
                session,
                reason="Created connection response",
                log_params={"response": response},
            )

        async with self.profile.session() as session:
            send_mediation_request = await conn_rec.metadata_get(
                session, MediationManager.SEND_REQ_AFTER_CONNECTION
            )
        if send_mediation_request:
            temp_mediation_mgr = MediationManager(self.profile)
            _, request = await temp_mediation_mgr.prepare_request(conn_rec.connection_id)
            responder = self.profile.inject(BaseResponder)
            await responder.send(request, connection_id=conn_rec.connection_id)

        return response

    async def accept_response(
        self,
        response: DIDXResponse,
        receipt: MessageReceipt,
    ) -> ConnRecord:
        """Accept a connection response under RFC 23 (DID exchange).

        Process a `DIDXResponse` message by looking up
        the connection request and setting up the pairwise connection.

        Args:
            response: The `DIDXResponse` to accept
            receipt: The message receipt

        Returns:
            The updated `ConnRecord` representing the connection

        Raises:
            DIDXManagerError: If there is no DID associated with the
                connection response
            DIDXManagerError: If the corresponding connection is not
                in the request-sent state

        """

        conn_rec = None
        if response._thread:
            # identify the request by the thread ID
            async with self.profile.session() as session:
                try:
                    conn_rec = await ConnRecord.retrieve_by_request_id(
                        session,
                        response._thread_id,
                        their_role=ConnRecord.Role.RESPONDER.rfc23,
                    )
                except StorageNotFoundError:
                    pass
                if not conn_rec:
                    try:
                        conn_rec = await ConnRecord.retrieve_by_request_id(
                            session,
                            response._thread_id,
                            their_role=ConnRecord.Role.RESPONDER.rfc160,
                        )
                    except StorageNotFoundError:
                        pass

        if not conn_rec and receipt.sender_did:
            # identify connection by the DID they used for us
            try:
                async with self.profile.session() as session:
                    conn_rec = await ConnRecord.retrieve_by_did(
                        session=session,
                        their_did=receipt.sender_did,
                        my_did=receipt.recipient_did,
                        their_role=ConnRecord.Role.RESPONDER.rfc23,
                    )
            except StorageNotFoundError:
                pass

        if not conn_rec:
            raise DIDXManagerError(
                "No corresponding connection request found",
                error_code=ProblemReportReason.RESPONSE_NOT_ACCEPTED.value,
            )

        if ConnRecord.State.get(conn_rec.state) is not ConnRecord.State.REQUEST:
            raise DIDXManagerError(
                "Cannot accept connection response for connection"
                f" in state: {conn_rec.state}"
            )

        their_did = response.did
        if response.did_doc_attach:
            async with self.profile.session() as session:
                wallet = session.inject(BaseWallet)
                conn_did_doc = await self.verify_diddoc(
                    wallet, response.did_doc_attach, conn_rec.invitation_key
                )
            # Special case: legacy DIDs were unqualified in response, qualified in doc
            if their_did and not their_did.startswith("did:"):
                did_to_check = f"did:sov:{their_did}"
            else:
                did_to_check = their_did

            if did_to_check != conn_did_doc["id"]:
                raise DIDXManagerError(
                    f"Connection DID {their_did} "
                    f"does not match DID doc id {conn_did_doc['id']}"
                )
            await self.store_did_document(conn_did_doc)
        else:
            if response.did is None:
                raise DIDXManagerError("No DID in response")

            if response.did_rotate_attach is None:
                raise DIDXManagerError(
                    "did_rotate~attach required if no signed doc attachment"
                )

            self._logger.debug("did_rotate~attach found; verifying signature")
            async with self.profile.session() as session:
                wallet = session.inject(BaseWallet)
                signed_did = await self.verify_rotate(
                    wallet, response.did_rotate_attach, conn_rec.invitation_key
                )
                if their_did != signed_did:
                    raise DIDXManagerError(
                        f"Connection DID {their_did} "
                        f"does not match signed DID rotate {signed_did}"
                    )

            self._logger.debug(
                "No DID Doc attachment in response; doc will be resolved from DID"
            )
            await self.record_keys_for_resolvable_did(response.did)

        conn_rec.their_did = their_did

        # The long format I sent has been acknowledged, use short form now.
        if LONG_PATTERN.match(conn_rec.my_did or ""):
            conn_rec.my_did = await self.long_did_peer_4_to_short(conn_rec.my_did)
        if LONG_PATTERN.match(conn_rec.their_did or ""):
            conn_rec.their_did = long_to_short(conn_rec.their_did)

        conn_rec.state = ConnRecord.State.RESPONSE.rfc160
        async with self.profile.session() as session:
            await conn_rec.save(session, reason="Accepted connection response")

        async with self.profile.session() as session:
            send_mediation_request = await conn_rec.metadata_get(
                session, MediationManager.SEND_REQ_AFTER_CONNECTION
            )
        if send_mediation_request:
            temp_mediation_mgr = MediationManager(self.profile)
            _record, request = await temp_mediation_mgr.prepare_request(
                conn_rec.connection_id
            )
            responder = self.profile.inject(BaseResponder)
            await responder.send(request, connection_id=conn_rec.connection_id)

        # create and send connection-complete message
        complete = DIDXComplete()
        complete.assign_thread_from(response)
        if conn_rec.connection_protocol == DIDEX_1_0:
            complete.assign_version("1.0")

        responder = self.profile.inject_or(BaseResponder)
        if responder:
            await responder.send_reply(complete, connection_id=conn_rec.connection_id)

            conn_rec.state = ConnRecord.State.COMPLETED.rfc160
            async with self.profile.session() as session:
                await conn_rec.save(session, reason="Sent connection complete")
                if session.settings.get("auto_disclose_features"):
                    discovery_mgr = V20DiscoveryMgr(self._profile)
                    await discovery_mgr.proactive_disclose_features(
                        connection_id=conn_rec.connection_id
                    )

        return conn_rec

    async def accept_complete(
        self,
        complete: DIDXComplete,
        receipt: MessageReceipt,
    ) -> ConnRecord:
        """Accept a connection complete message under RFC 23 (DID exchange).

        Process a `DIDXComplete` message by looking up
        the connection record and marking the exchange complete.

        Args:
            complete: The `DIDXComplete` to accept
            receipt: The message receipt

        Returns:
            The updated `ConnRecord` representing the connection

        Raises:
            DIDXManagerError: If the corresponding connection does not exist
                or is not in the response-sent state

        """
        conn_rec = None

        # identify the request by the thread ID
        async with self.profile.session() as session:
            try:
                conn_rec = await ConnRecord.retrieve_by_request_id(
                    session,
                    complete._thread_id,
                    their_role=ConnRecord.Role.REQUESTER.rfc23,
                )
            except StorageNotFoundError:
                pass

            if not conn_rec:
                try:
                    conn_rec = await ConnRecord.retrieve_by_request_id(
                        session,
                        complete._thread_id,
                        their_role=ConnRecord.Role.REQUESTER.rfc160,
                    )
                except StorageNotFoundError:
                    pass

        if not conn_rec:
            raise DIDXManagerError(
                "No corresponding connection request found",
                error_code=ProblemReportReason.COMPLETE_NOT_ACCEPTED.value,
            )

        if LONG_PATTERN.match(conn_rec.my_did or ""):
            conn_rec.my_did = await self.long_did_peer_4_to_short(conn_rec.my_did)
        if LONG_PATTERN.match(conn_rec.their_did or ""):
            conn_rec.their_did = long_to_short(conn_rec.their_did)

        conn_rec.state = ConnRecord.State.COMPLETED.rfc160
        async with self.profile.session() as session:
            await conn_rec.save(session, reason="Received connection complete")
            if session.settings.get("auto_disclose_features"):
                discovery_mgr = V20DiscoveryMgr(self._profile)
                await discovery_mgr.proactive_disclose_features(
                    connection_id=conn_rec.connection_id
                )

        return conn_rec

    async def reject(
        self,
        conn_rec: ConnRecord,
        *,
        reason: Optional[str] = None,
    ) -> DIDXProblemReport:
        """Abandon an existing DID exchange."""
        state_to_reject_code = {
            ConnRecord.State.INVITATION.rfc23
            + "-received": ProblemReportReason.INVITATION_NOT_ACCEPTED,
            ConnRecord.State.REQUEST.rfc23
            + "-received": ProblemReportReason.REQUEST_NOT_ACCEPTED,
        }
        code = state_to_reject_code.get(conn_rec.rfc23_state)
        if not code:
            raise DIDXManagerError(
                f"Cannot reject connection in state: {conn_rec.rfc23_state}"
            )

        async with self.profile.session() as session:
            await conn_rec.abandon(session, reason=reason)

        report = DIDXProblemReport(
            description={
                "code": code.value,
                "en": reason or "DID exchange rejected",
            },
        )
        if conn_rec.connection_protocol == DIDEX_1_0:
            report.assign_version("1.0")

        # TODO Delete the record?
        return report

    async def receive_problem_report(
        self,
        conn_rec: ConnRecord,
        report: DIDXProblemReport,
    ):
        """Receive problem report."""
        if not report.description:
            raise DIDXManagerError("Missing description in problem report")

        if report.description.get("code") in {
            reason.value for reason in ProblemReportReason
        }:
            self._logger.info("Problem report indicates connection is abandoned")
            async with self.profile.session() as session:
                await conn_rec.abandon(
                    session,
                    reason=report.description.get("en"),
                )
        else:
            raise DIDXManagerError(
                f"Received unrecognized problem report: {report.description}"
            )

    async def verify_diddoc(
        self,
        wallet: BaseWallet,
        attached: AttachDecorator,
        invi_key: Optional[str] = None,
    ) -> dict:
        """Verify DIDDoc attachment and return signed data."""
        signed_diddoc_bytes = attached.data.signed
        if not signed_diddoc_bytes:
            raise DIDXManagerError("DID doc attachment is not signed.")
        if not await attached.data.verify(wallet, invi_key):
            raise DIDXManagerError("DID doc attachment signature failed verification")

        return json.loads(signed_diddoc_bytes.decode())

    async def verify_rotate(
        self,
        wallet: BaseWallet,
        attached: AttachDecorator,
        invi_key: Optional[str] = None,
    ) -> str:
        """Verify a signed DID rotate attachment and return did."""
        signed_diddoc_bytes = attached.data.signed
        if not signed_diddoc_bytes:
            raise DIDXManagerError("DID rotate attachment is not signed.")
        if not await attached.data.verify(wallet, invi_key):
            raise DIDXManagerError("DID rotate attachment signature failed verification")

        return signed_diddoc_bytes.decode()

    async def manager_error_to_problem_report(
        self,
        e: DIDXManagerError,
        message: Union[DIDXRequest, DIDXResponse],
        message_receipt,
    ) -> tuple[DIDXProblemReport, Sequence[ConnectionTarget]]:
        """Convert DIDXManagerError to problem report."""
        self._logger.exception("Error receiving RFC 23 connection request")
        targets = None
        report = None
        if e.error_code:
            report = DIDXProblemReport(
                description={"en": e.message, "code": e.error_code}
            )
            report.assign_thread_from(message)
            report.assign_version_from(message)
            if message.did_doc_attach:
                try:
                    # convert diddoc attachment to diddoc...
                    async with self.profile.session() as session:
                        wallet = session.inject(BaseWallet)
                        conn_did_doc = await self.verify_diddoc(
                            wallet, message.did_doc_attach
                        )
                    # get the connection targets...
                    targets = self.diddoc_connection_targets(
                        conn_did_doc,
                        message_receipt.recipient_verkey,
                    )
                except DIDXManagerError:
                    self._logger.exception("Error parsing DIDDoc for problem report")

        return report, targets



from datetime import datetime
from unittest import IsolatedAsyncioTestCase, mock

import pytest

from ....did.did_key import DIDKey
from ....utils.testing import create_test_profile
from ....wallet.base import BaseWallet
from ....wallet.key_type import BLS12381G2, ED25519
from ...ld_proofs import (
    BbsBlsSignature2020,
    Ed25519Signature2018,
    Ed25519Signature2020,
    WalletKeyPair,
)
from ...ld_proofs.error import LinkedDataProofException
from ...tests.document_loader import custom_document_loader
from ...vc_ld import create_presentation, derive_credential
from ...vc_ld import issue_vc as issue
from ...vc_ld import sign_presentation, verify_credential, verify_presentation
from .test_credential import (
    CREDENTIAL_ISSUED,
    CREDENTIAL_ISSUED_2020,
    CREDENTIAL_ISSUED_BBS,
    CREDENTIAL_TEMPLATE,
    CREDENTIAL_TEMPLATE_2020,
    CREDENTIAL_TEMPLATE_BBS,
    CREDENTIAL_VERIFIED,
    CREDENTIAL_VERIFIED_2020,
    PRESENTATION_SIGNED,
    PRESENTATION_UNSIGNED,
)


class TestLinkedDataVerifiableCredential(IsolatedAsyncioTestCase):
    test_seed = "testseed000000000000000000000001"

    async def asyncSetUp(self):
        self.profile = await create_test_profile()
        async with self.profile.session() as session:
            wallet = session.inject(BaseWallet)
            self.ed25519_key_info = await wallet.create_signing_key(
                key_type=ED25519, seed=self.test_seed
            )
            self.ed25519_verification_method = DIDKey.from_public_key_b58(
                self.ed25519_key_info.verkey, ED25519
            ).key_id

            self.bls12381g2_key_info = await wallet.create_signing_key(
                key_type=BLS12381G2, seed=self.test_seed
            )

            self.bls12381g2_verification_method = DIDKey.from_public_key_b58(
                self.bls12381g2_key_info.verkey, BLS12381G2
            ).key_id

        self.presentation_challenge = "2b1bbff6-e608-4368-bf84-67471b27e41c"

    async def test_issue_Ed25519Signature2018(self):
        # Use different key pair and suite for signing and verification
        # as during verification a lot of information can be extracted
        # from the proof / document
        suite = Ed25519Signature2018(
            verification_method=self.ed25519_verification_method,
            key_pair=WalletKeyPair(
                profile=self.profile,
                key_type=ED25519,
                public_key_base58=self.ed25519_key_info.verkey,
            ),
            date=datetime.strptime("2019-12-11T03:50:55Z", "%Y-%m-%dT%H:%M:%SZ"),
        )

        issued = await issue(
            credential=CREDENTIAL_TEMPLATE,
            suite=suite,
            document_loader=custom_document_loader,
        )

        assert issued == CREDENTIAL_ISSUED

    async def test_issue_Ed25519Signature2020(self):
        # Use different key pair and suite for signing and verification
        # as during verification a lot of information can be extracted
        # from the proof / document
        suite = Ed25519Signature2020(
            verification_method=self.ed25519_verification_method,
            key_pair=WalletKeyPair(
                profile=self.profile,
                key_type=ED25519,
                public_key_base58=self.ed25519_key_info.verkey,
            ),
            date=datetime.strptime("2019-12-11T03:50:55Z", "%Y-%m-%dT%H:%M:%SZ"),
        )

        issued = await issue(
            credential=CREDENTIAL_TEMPLATE_2020,
            suite=suite,
            document_loader=custom_document_loader,
        )

        assert issued == CREDENTIAL_ISSUED_2020

    async def test_issue_x_invalid_credential_structure(self):
        credential = CREDENTIAL_TEMPLATE.copy()
        credential.pop("issuer")

        with self.assertRaises(LinkedDataProofException) as context:
            await issue(
                credential=credential,
                suite=mock.MagicMock(),
                document_loader=mock.MagicMock(),
            )
        assert "invalid structure" in str(context.exception)

    async def test_derive_x_invalid_credential_structure(self):
        credential = CREDENTIAL_TEMPLATE.copy()
        credential.pop("issuer")

        with self.assertRaises(LinkedDataProofException) as context:
            await derive_credential(
                credential=credential,
                reveal_document=mock.MagicMock(),
                suite=mock.MagicMock(),
                document_loader=mock.MagicMock(),
            )
        assert "invalid structure" in str(context.exception)

    async def test_verify_Ed25519Signature2018(self):
        # Verification requires lot less input parameters
        suite = Ed25519Signature2018(
            key_pair=WalletKeyPair(profile=self.profile, key_type=ED25519),
        )
        verified = await verify_credential(
            credential=CREDENTIAL_ISSUED,
            suites=[suite],
            document_loader=custom_document_loader,
        )

        assert verified == CREDENTIAL_VERIFIED

    async def test_verify_Ed25519Signature2020(self):
        # Verification requires lot less input parameters
        suite = Ed25519Signature2020(
            key_pair=WalletKeyPair(profile=self.profile, key_type=ED25519),
        )
        verified = await verify_credential(
            credential=CREDENTIAL_ISSUED_2020,
            suites=[suite],
            document_loader=custom_document_loader,
        )

        assert verified == CREDENTIAL_VERIFIED_2020

    async def test_verify_x_invalid_credential_structure(self):
        credential = CREDENTIAL_ISSUED.copy()
        credential.pop("issuer")

        result = await verify_credential(
            credential=credential,
            suites=[],
            document_loader=mock.MagicMock(),
        )

        assert not result.verified
        assert "invalid structure" in str(result.errors[0])

    @pytest.mark.ursa_bbs_signatures
    async def test_issue_BbsBlsSignature2020(self):
        # Use different key pair and suite for signing and verification
        # as during verification a lot of information can be extracted
        # from the proof / document
        suite = BbsBlsSignature2020(
            verification_method=self.bls12381g2_verification_method,
            key_pair=WalletKeyPair(
                profile=self.profile,
                key_type=BLS12381G2,
                public_key_base58=self.bls12381g2_key_info.verkey,
            ),
            date=datetime.strptime("2019-12-11T03:50:55Z", "%Y-%m-%dT%H:%M:%SZ"),
        )

        issued = await issue(
            credential=CREDENTIAL_TEMPLATE_BBS,
            suite=suite,
            document_loader=custom_document_loader,
        )

        assert issued

        result = await verify_credential(
            credential=issued, suites=[suite], document_loader=custom_document_loader
        )

        assert result.verified is not None

    async def test_verify_BbsBlsSignature2020(self):
        # Verification requires lot less input parameters
        suite = BbsBlsSignature2020(
            key_pair=WalletKeyPair(profile=self.profile, key_type=BLS12381G2),
        )
        result = await verify_credential(
            credential=CREDENTIAL_ISSUED_BBS,
            suites=[suite],
            document_loader=custom_document_loader,
        )

        assert result.verified

    async def test_create_presentation(self):
        credential = CREDENTIAL_ISSUED.copy()
        credential.pop("issuer")

        with self.assertRaises(LinkedDataProofException) as context:
            await create_presentation(credentials=[credential])
        assert "Not all credentials have a valid structure" in str(context.exception)

    async def test_create_presentation_x_invalid_credential_structures(self):
        unsigned_presentation = await create_presentation(credentials=[CREDENTIAL_ISSUED])

        suite = Ed25519Signature2018(
            verification_method=self.ed25519_verification_method,
            key_pair=WalletKeyPair(
                profile=self.profile,
                key_type=ED25519,
                public_key_base58=self.ed25519_key_info.verkey,
            ),
            date=datetime.strptime("2020-12-11T03:50:55Z", "%Y-%m-%dT%H:%M:%SZ"),
        )

        assert unsigned_presentation == PRESENTATION_UNSIGNED

        presentation = await sign_presentation(
            presentation=unsigned_presentation,
            suite=suite,
            document_loader=custom_document_loader,
            challenge=self.presentation_challenge,
        )

        assert presentation == PRESENTATION_SIGNED

        # test id property
        assert "id" in await create_presentation(
            credentials=[CREDENTIAL_ISSUED],
            presentation_id="https://presentation_id.com",
        )

    @pytest.mark.ursa_bbs_signatures
    async def test_sign_presentation_bbsbls(self):
        unsigned_presentation = await create_presentation(credentials=[CREDENTIAL_ISSUED])

        suite = BbsBlsSignature2020(
            verification_method=self.bls12381g2_verification_method,
            key_pair=WalletKeyPair(
                profile=self.profile,
                key_type=BLS12381G2,
                public_key_base58=self.bls12381g2_key_info.verkey,
            ),
            date=datetime.strptime("2020-12-11T03:50:55Z", "%Y-%m-%dT%H:%M:%SZ"),
        )

        assert unsigned_presentation == PRESENTATION_UNSIGNED
        unsigned_presentation["@context"].append("https://w3id.org/security/bbs/v1")
        _ = await sign_presentation(
            presentation=unsigned_presentation,
            suite=suite,
            document_loader=custom_document_loader,
            challenge=self.presentation_challenge,
        )

    async def test_verify_presentation(self):
        suite = Ed25519Signature2018(
            key_pair=WalletKeyPair(profile=self.profile, key_type=ED25519),
        )
        verification_result = await verify_presentation(
            presentation=PRESENTATION_SIGNED,
            challenge=self.presentation_challenge,
            suites=[suite],
            document_loader=custom_document_loader,
        )

        assert verification_result.verified

    async def test_verify_presentation_x_no_purpose_challenge(self):
        verification_result = await verify_presentation(
            presentation=PRESENTATION_SIGNED,
            suites=[],
            document_loader=custom_document_loader,
        )

        assert not verification_result.verified
        assert 'A "challenge" param is required for AuthenticationProofPurpose' in str(
            verification_result.errors[0]
        )

    async def test_sign_presentation_x_no_purpose_challenge(self):
        with self.assertRaises(LinkedDataProofException) as context:
            await sign_presentation(
                presentation=PRESENTATION_UNSIGNED,
                suite=mock.MagicMock(),
                document_loader=mock.MagicMock(),
            )
        assert 'A "challenge" param is required' in str(context.exception)

    async def test_verify_x_no_proof(self):
        presentation = PRESENTATION_SIGNED.copy()
        presentation.pop("proof")

        verification_result = await verify_presentation(
            presentation=presentation,
            challenge=self.presentation_challenge,
            suites=[],
            document_loader=custom_document_loader,
        )

        assert not verification_result.verified
        assert 'presentation must contain "proof"' in str(verification_result.errors[0])



"""The Conductor.

The conductor is responsible for coordinating messages that are received
over the network, communicating with the ledger, passing messages to handlers,
instantiating concrete implementations of required modules and storing data in the
wallet.

"""

import asyncio
import hashlib
import json
import logging
from typing import Optional

from packaging import version as package_version
from qrcode import QRCode

from ..admin.base_server import BaseAdminServer
from ..admin.server import AdminResponder, AdminServer
from ..commands.upgrade import add_version_record, get_upgrade_version_list, upgrade
from ..config.default_context import ContextBuilder, DefaultContextBuilder
from ..config.injection_context import InjectionContext
from ..config.ledger import (
    get_genesis_transactions,
    ledger_config,
    load_multiple_genesis_transactions_from_config,
)
from ..config.logging import LoggingConfigurator
from ..config.provider import ClassProvider
from ..config.wallet import wallet_config
from ..connections.base_manager import BaseConnectionManager, BaseConnectionManagerError
from ..core.profile import Profile
from ..indy.verifier import IndyVerifier
from ..ledger.base import BaseLedger
from ..ledger.error import LedgerConfigError, LedgerTransactionError
from ..ledger.multiple_ledger.base_manager import (
    BaseMultipleLedgerManager,
    MultipleLedgerManagerError,
)
from ..ledger.multiple_ledger.ledger_requests_executor import IndyLedgerRequestsExecutor
from ..ledger.multiple_ledger.manager_provider import MultiIndyLedgerManagerProvider
from ..messaging.responder import BaseResponder
from ..multitenant.base import BaseMultitenantManager
from ..multitenant.manager_provider import MultitenantManagerProvider
from ..protocols.coordinate_mediation.mediation_invite_store import MediationInviteStore
from ..protocols.coordinate_mediation.v1_0.manager import MediationManager
from ..protocols.coordinate_mediation.v1_0.route_manager import RouteManager
from ..protocols.coordinate_mediation.v1_0.route_manager_provider import (
    RouteManagerProvider,
)
from ..protocols.out_of_band.v1_0.manager import OutOfBandManager
from ..protocols.out_of_band.v1_0.messages.invitation import HSProto, InvitationMessage
from ..storage.base import BaseStorage
from ..storage.error import StorageNotFoundError
from ..storage.record import StorageRecord
from ..storage.type import (
    RECORD_TYPE_ACAPY_STORAGE_TYPE,
    STORAGE_TYPE_VALUE_ANONCREDS,
    STORAGE_TYPE_VALUE_ASKAR,
)
from ..transport.inbound.manager import InboundTransportManager
from ..transport.inbound.message import InboundMessage
from ..transport.outbound.base import OutboundDeliveryError
from ..transport.outbound.manager import OutboundTransportManager, QueuedOutboundMessage
from ..transport.outbound.message import OutboundMessage
from ..transport.outbound.status import OutboundSendStatus
from ..transport.wire_format import BaseWireFormat
from ..utils.profiles import get_subwallet_profiles_from_storage
from ..utils.stats import Collector
from ..utils.task_queue import CompletedTask, TaskQueue
from ..vc.ld_proofs.document_loader import DocumentLoader
from ..version import RECORD_TYPE_ACAPY_VERSION, __version__
from ..wallet.anoncreds_upgrade import upgrade_wallet_to_anoncreds_if_requested
from ..wallet.did_info import DIDInfo
from ..wallet.singletons import IsAnonCredsSingleton
from .dispatcher import Dispatcher
from .error import ProfileError, StartupError
from .oob_processor import OobMessageProcessor
from .util import SHUTDOWN_EVENT_TOPIC, STARTUP_EVENT_TOPIC

LOGGER = logging.getLogger(__name__)
# Refer ACA-Py issue #2197
# When the from version is not found
DEFAULT_ACAPY_VERSION = "v0.7.5"


class Conductor:
    """Conductor class.

    Class responsible for initializing concrete implementations
    of our required interfaces and routing inbound and outbound message data.
    """

    def __init__(self, context_builder: ContextBuilder) -> None:
        """Initialize an instance of Conductor.

        Args:
            inbound_transports: Configuration for inbound transports
            outbound_transports: Configuration for outbound transports
            settings: Dictionary of various settings
            context_builder: Context builder for the conductor

        """
        self.admin_server = None
        self.context_builder = context_builder
        self.dispatcher: Optional[Dispatcher] = None
        self.inbound_transport_manager: Optional[InboundTransportManager] = None
        self.outbound_transport_manager: Optional[OutboundTransportManager] = None
        self.root_profile: Optional[Profile] = None
        self.setup_public_did: Optional[DIDInfo] = None

    force_agent_anoncreds = False

    @property
    def context(self) -> InjectionContext:
        """Accessor for the injection context."""
        assert self.root_profile, "root_profile is not set"
        return self.root_profile.context

    async def setup(self):
        """Initialize the global request context."""
        LOGGER.debug("Starting setup of the Conductor")

        context = await self.context_builder.build_context()
        LOGGER.debug("Context built successfully")

        if self.force_agent_anoncreds:
            LOGGER.debug(
                "Force agent anoncreds is enabled. "
                "Setting wallet type to 'askar-anoncreds'."
            )
            context.settings.set_value("wallet.type", "askar-anoncreds")

        # Fetch genesis transactions if necessary
        if context.settings.get("ledger.ledger_config_list"):
            LOGGER.debug(
                "Ledger config list found. Loading multiple genesis transactions"
            )
            await load_multiple_genesis_transactions_from_config(context.settings)
        if (
            context.settings.get("ledger.genesis_transactions")
            or context.settings.get("ledger.genesis_file")
            or context.settings.get("ledger.genesis_url")
        ):
            LOGGER.debug(
                "Genesis transactions/configurations found. Fetching genesis transactions"
            )
            await get_genesis_transactions(context.settings)

        # Configure the root profile
        LOGGER.debug("Configuring the root profile and setting up public DID")
        self.root_profile, self.setup_public_did = await wallet_config(context)
        context = self.root_profile.context
        LOGGER.debug("Root profile configured successfully")

        # Multiledger Setup
        ledger_config_list = context.settings.get("ledger.ledger_config_list")
        if ledger_config_list and len(ledger_config_list) > 0:
            LOGGER.debug("Setting up multiledger manager")
            context.injector.bind_provider(
                BaseMultipleLedgerManager,
                MultiIndyLedgerManagerProvider(self.root_profile),
            )
            if not context.settings.get("ledger.genesis_transactions"):
                ledger = context.injector.inject(BaseLedger)
                LOGGER.debug(
                    "Ledger backend: %s, Profile backend: %s",
                    ledger.BACKEND_NAME,
                    self.root_profile.BACKEND_NAME,
                )
                if (
                    self.root_profile.BACKEND_NAME == "askar"
                    and ledger.BACKEND_NAME == "indy-vdr"
                ):
                    LOGGER.debug("Binding IndyCredxVerifier for 'askar' backend.")
                    context.injector.bind_provider(
                        IndyVerifier,
                        ClassProvider(
                            "acapy_agent.indy.credx.verifier.IndyCredxVerifier",
                            self.root_profile,
                        ),
                    )
                elif (
                    self.root_profile.BACKEND_NAME == "askar-anoncreds"
                    and ledger.BACKEND_NAME == "indy-vdr"
                ):
                    LOGGER.debug(
                        "Binding IndyCredxVerifier for 'askar-anoncreds' backend."
                    )
                    context.injector.bind_provider(
                        IndyVerifier,
                        ClassProvider(
                            "acapy_agent.anoncreds.credx.verifier.IndyCredxVerifier",
                            self.root_profile,
                        ),
                    )
                else:
                    LOGGER.error("Unsupported ledger backend for multiledger setup.")
                    raise MultipleLedgerManagerError(
                        "Multiledger is supported only for Indy SDK or Askar "
                        "[Indy VDR] profile"
                    )
        context.injector.bind_instance(
            IndyLedgerRequestsExecutor, IndyLedgerRequestsExecutor(self.root_profile)
        )

        # Configure the ledger
        ledger_configured = await ledger_config(
            self.root_profile, self.setup_public_did and self.setup_public_did.did
        )
        if not ledger_configured:
            LOGGER.info("No ledger configured.")
        else:
            LOGGER.info("Ledger configured successfully.")

        if not context.settings.get("transport.disabled"):
            # Register all inbound transports if enabled
            LOGGER.debug("Transport not disabled. Setting up inbound transports.")
            self.inbound_transport_manager = InboundTransportManager(
                self.root_profile, self.inbound_message_router, self.handle_not_returned
            )
            await self.inbound_transport_manager.setup()
            context.injector.bind_instance(
                InboundTransportManager, self.inbound_transport_manager
            )
            LOGGER.debug("Inbound transports registered successfully.")

            # Register all outbound transports
            LOGGER.debug("Setting up outbound transports.")
            self.outbound_transport_manager = OutboundTransportManager(
                self.root_profile, self.handle_not_delivered
            )
            await self.outbound_transport_manager.setup()
            LOGGER.debug("Outbound transports registered successfully.")

        # Initialize dispatcher
        LOGGER.debug("Initializing dispatcher.")
        self.dispatcher = Dispatcher(self.root_profile)
        await self.dispatcher.setup()
        LOGGER.debug("Dispatcher initialized successfully.")

        wire_format = context.inject_or(BaseWireFormat)
        if wire_format and hasattr(wire_format, "task_queue"):
            wire_format.task_queue = self.dispatcher.task_queue
            LOGGER.debug("Wire format task queue bound to dispatcher.")

        # Bind manager for multitenancy related tasks
        if context.settings.get("multitenant.enabled"):
            LOGGER.debug("Multitenant is enabled. Binding MultitenantManagerProvider.")
            context.injector.bind_provider(
                BaseMultitenantManager, MultitenantManagerProvider(self.root_profile)
            )

        # Bind route manager provider
        LOGGER.debug("Binding RouteManagerProvider.")
        context.injector.bind_provider(
            RouteManager, RouteManagerProvider(self.root_profile)
        )

        # Bind OobMessageProcessor to be able to receive and process unencrypted messages
        LOGGER.debug("Binding OobMessageProcessor.")
        context.injector.bind_instance(
            OobMessageProcessor,
            OobMessageProcessor(inbound_message_router=self.inbound_message_router),
        )

        # Bind default PyLD document loader
        LOGGER.debug("Binding default DocumentLoader.")
        context.injector.bind_instance(DocumentLoader, DocumentLoader(self.root_profile))

        # Admin API
        if context.settings.get("admin.enabled"):
            LOGGER.debug("Admin API is enabled. Attempting to register admin server.")
            try:
                admin_host = context.settings.get("admin.host", "0.0.0.0")
                admin_port = context.settings.get("admin.port", "80")
                self.admin_server = AdminServer(
                    admin_host,
                    admin_port,
                    context,
                    self.root_profile,
                    self.outbound_message_router,
                    self.webhook_router,
                    self.stop,
                    self.dispatcher.task_queue,
                    self.get_stats,
                )
                context.injector.bind_instance(BaseAdminServer, self.admin_server)
                LOGGER.debug("Admin server registered on %s:%s", admin_host, admin_port)
            except Exception:
                LOGGER.exception("Unable to register admin server.")
                raise

        # Fetch stats collector, if any
        collector = context.inject_or(Collector)
        if collector:
            LOGGER.debug("Stats collector found. Wrapping methods for collection.")
            # add stats to our own methods
            collector.wrap(
                self,
                (
                    # "inbound_message_router",
                    "outbound_message_router",
                    # "create_inbound_session",
                ),
            )
            # at the class level (!) should not be performed multiple times
            collector.wrap(
                BaseConnectionManager,
                (
                    # "get_connection_targets",
                    "fetch_did_document",
                    "find_inbound_connection",
                ),
            )
            LOGGER.debug("Methods wrapped with stats collector.")

    async def start(self) -> None:
        """Start the agent."""
        LOGGER.debug("Starting the Conductor agent.")
        assert self.root_profile, "root_profile is not set"
        context = self.root_profile.context
        await self.check_for_valid_wallet_type(self.root_profile)
        LOGGER.debug("Wallet type validated.")

        if not context.settings.get("transport.disabled"):
            # Start up transports if enabled
            try:
                LOGGER.debug("Transport not disabled. Starting inbound transports.")
                await self.inbound_transport_manager.start()
                LOGGER.debug("Inbound transports started successfully.")
            except Exception:
                LOGGER.exception("Unable to start inbound transports.")
                raise
            try:
                LOGGER.debug("Starting outbound transports.")
                await self.outbound_transport_manager.start()
                LOGGER.debug("Outbound transports started successfully.")
            except Exception:
                LOGGER.exception("Unable to start outbound transports.")
                raise

        # Start up Admin server
        if self.admin_server:
            LOGGER.debug("Admin server present. Starting admin server.")
            try:
                await self.admin_server.start()
                LOGGER.debug("Admin server started successfully.")
            except Exception:
                LOGGER.exception("Unable to start administration API.")
            # Make admin responder available during message parsing
            # This allows webhooks to be called when a connection is marked active,
            # for example
            responder = AdminResponder(
                self.root_profile,
                self.admin_server.outbound_message_router,
            )
            context.injector.bind_instance(BaseResponder, responder)
            LOGGER.debug("Admin responder bound to injector.")

        # Get agent label
        default_label = context.settings.get("default_label")
        LOGGER.debug("Agent label: %s", default_label)

        if context.settings.get("transport.disabled"):
            LoggingConfigurator.print_banner(
                default_label,
                None,
                None,
                self.setup_public_did and self.setup_public_did.did,
                self.admin_server,
            )
        else:
            LoggingConfigurator.print_banner(
                default_label,
                self.inbound_transport_manager.registered_transports,
                self.outbound_transport_manager.registered_transports,
                self.setup_public_did and self.setup_public_did.did,
                self.admin_server,
            )

        LoggingConfigurator.print_notices(context.settings)

        # record ACA-Py version in Wallet, if needed
        from_version_storage = None
        from_version = None
        agent_version = f"v{__version__}"
        LOGGER.debug("Recording ACA-Py version in wallet if needed.")
        async with self.root_profile.session() as session:
            storage: BaseStorage = session.context.inject(BaseStorage)
            try:
                record = await storage.find_record(
                    type_filter=RECORD_TYPE_ACAPY_VERSION,
                    tag_query={},
                )
                from_version_storage = record.value
                LOGGER.info(
                    "Existing acapy_version storage record found, version set to %s",
                    from_version_storage,
                )
            except StorageNotFoundError:
                LOGGER.info("Wallet version storage record not found.")

        from_version_config = self.root_profile.settings.get("upgrade.from_version")
        force_upgrade_flag = (
            self.root_profile.settings.get("upgrade.force_upgrade") or False
        )
        LOGGER.debug(
            "Force upgrade flag: %s, From version config: %s",
            force_upgrade_flag,
            from_version_config,
        )

        if force_upgrade_flag and from_version_config:
            if from_version_storage:
                if package_version.parse(from_version_storage) > package_version.parse(
                    from_version_config
                ):
                    from_version = from_version_config
                else:
                    from_version = from_version_storage
            else:
                from_version = from_version_config
            LOGGER.debug(
                "Determined from_version based on force_upgrade: %s", from_version
            )
        else:
            from_version = from_version_storage or from_version_config
            LOGGER.debug("Determined from_version: %s", from_version)

        if not from_version:
            LOGGER.info(
                "No upgrade from version was found from wallet or via"
                " --from-version startup argument. Defaulting to %s.",
                DEFAULT_ACAPY_VERSION,
            )
            from_version = DEFAULT_ACAPY_VERSION
            self.root_profile.settings.set_value("upgrade.from_version", from_version)
            LOGGER.debug("Set upgrade.from_version to default: %s", from_version)

        config_available_list = get_upgrade_version_list(
            config_path=self.root_profile.settings.get("upgrade.config_path"),
            from_version=from_version,
        )
        LOGGER.debug("Available upgrade versions: %s", config_available_list)

        if len(config_available_list) >= 1:
            LOGGER.info("Upgrade configurations available. Initiating upgrade.")
            await upgrade(profile=self.root_profile)
        elif not (from_version_storage and from_version_storage == agent_version):
            LOGGER.debug("No upgrades needed. Adding version record.")
            await add_version_record(profile=self.root_profile, version=agent_version)

        # Create a static connection for use by the test-suite
        if context.settings.get("debug.test_suite_endpoint"):
            LOGGER.debug(
                "Test suite endpoint configured. "
                "Creating static connection for test suite."
            )
            mgr = BaseConnectionManager(self.root_profile)
            their_endpoint = context.settings["debug.test_suite_endpoint"]
            _, _, test_conn = await mgr.create_static_connection(
                my_seed=hashlib.sha256(b"aries-protocol-test-subject").digest(),
                their_seed=hashlib.sha256(b"aries-protocol-test-suite").digest(),
                their_endpoint=their_endpoint,
                alias="test-suite",
            )
            LOGGER.info(
                "Created static connection for test suite\n"
                " - My DID: %s\n"
                " - Their DID: %s\n"
                " - Their endpoint: %s\n",
                test_conn.my_did,
                test_conn.their_did,
                their_endpoint,
            )
            del mgr
            LOGGER.debug("Static connection for test suite created and manager deleted.")

        # Clear default mediator
        if context.settings.get("mediation.clear"):
            LOGGER.debug("Mediation clear flag set. Clearing default mediator.")
            mediation_mgr = MediationManager(self.root_profile)
            await mediation_mgr.clear_default_mediator()
            LOGGER.info("Default mediator cleared.")

        # Set default mediator by id
        default_mediator_id = context.settings.get("mediation.default_id")
        if default_mediator_id:
            LOGGER.debug("Setting default mediator to ID: %s", default_mediator_id)
            mediation_mgr = MediationManager(self.root_profile)
            try:
                await mediation_mgr.set_default_mediator_by_id(default_mediator_id)
                LOGGER.info("Default mediator set to %s", default_mediator_id)
            except Exception:
                LOGGER.exception("Error updating default mediator.")

        # Print an invitation to the terminal
        if context.settings.get("debug.print_invitation"):
            LOGGER.debug(
                "Debug flag for printing invitation is set. Creating invitation."
            )
            try:
                mgr = OutOfBandManager(self.root_profile)
                invi_rec = await mgr.create_invitation(
                    my_label=context.settings.get("debug.invite_label"),
                    public=context.settings.get("debug.invite_public", False),
                    multi_use=context.settings.get("debug.invite_multi_use", False),
                    hs_protos=[HSProto.RFC23],
                    metadata=json.loads(
                        context.settings.get("debug.invite_metadata_json", "{}")
                    ),
                )
                base_url = context.settings.get("invite_base_url")
                invite_url = invi_rec.invitation.to_url(base_url)
                LOGGER.info("Invitation URL:\n%s", invite_url)
                qr = QRCode(border=1)
                qr.add_data(invite_url)
                qr.print_ascii(invert=True)
                del mgr
            except Exception:
                LOGGER.exception("Error creating invitation.")

        # mediation connection establishment
        provided_invite: str = context.settings.get("mediation.invite")
        LOGGER.debug("Mediation invite provided: %s", provided_invite)

        try:
            async with self.root_profile.session() as session:
                invite_store = MediationInviteStore(session.context.inject(BaseStorage))
                mediation_invite_record = await invite_store.get_mediation_invite_record(
                    provided_invite
                )
                LOGGER.debug("Mediation invite record retrieved successfully.")
        except Exception:
            LOGGER.exception("Error retrieving mediator invitation.")
            mediation_invite_record = None

        # Accept mediation invitation if one was specified or stored
        if mediation_invite_record is not None:
            LOGGER.debug(
                "Mediation invite record found. "
                "Attempting to accept mediation invitation."
            )
            try:
                if not mediation_invite_record.used:
                    # clear previous mediator configuration before establishing a new one
                    LOGGER.debug(
                        "Mediation invite not used. "
                        "Clearing default mediator before accepting new invite."
                    )
                    await MediationManager(self.root_profile).clear_default_mediator()

                    mgr = OutOfBandManager(self.root_profile)
                    LOGGER.debug("Receiving mediation invitation.")
                    record = await mgr.receive_invitation(
                        invitation=InvitationMessage.from_url(
                            mediation_invite_record.invite
                        ),
                        auto_accept=True,
                    )
                    async with self.root_profile.session() as session:
                        await MediationInviteStore(
                            session.context.inject(BaseStorage)
                        ).mark_default_invite_as_used()
                        LOGGER.debug("Marked mediation invite as used.")

                        await record.metadata_set(
                            session, MediationManager.SEND_REQ_AFTER_CONNECTION, True
                        )
                        await record.metadata_set(
                            session, MediationManager.SET_TO_DEFAULT_ON_GRANTED, True
                        )
                        LOGGER.debug("Set mediation metadata after connection.")

                    LOGGER.info("Attempting to connect to mediator...")
                    del mgr
                    LOGGER.debug("Mediation manager deleted after setting up mediator.")
            except Exception:
                LOGGER.exception("Error accepting mediation invitation.")

        try:
            LOGGER.debug("Checking for wallet upgrades in progress.")
            await self.check_for_wallet_upgrades_in_progress()
            LOGGER.debug("Wallet upgrades check completed.")
        except Exception:
            LOGGER.exception(
                "An exception was caught while checking for wallet upgrades in progress."
            )

        # Ensure anoncreds wallet is added to singleton (avoids unnecessary upgrade check)
        if self.root_profile.settings.get("wallet.type") == "askar-anoncreds":
            IsAnonCredsSingleton().set_wallet(self.root_profile.name)

        # notify protocols of startup status
        LOGGER.debug("Notifying protocols of startup status.")
        await self.root_profile.notify(STARTUP_EVENT_TOPIC, {})
        LOGGER.debug("Startup notification sent.")

        LOGGER.info("Listening...")

    async def stop(self, timeout=1.0):
        """Stop the agent."""
        LOGGER.info("Stopping the Conductor agent.")
        # notify protocols that we are shutting down
        if self.root_profile:
            LOGGER.debug("Notifying protocols of shutdown.")
            await self.root_profile.notify(SHUTDOWN_EVENT_TOPIC, {})
            LOGGER.debug("Shutdown notification sent.")

        shutdown = TaskQueue()
        if self.dispatcher:
            LOGGER.debug("Initiating shutdown of dispatcher.")
            shutdown.run(self.dispatcher.complete())
        if self.admin_server:
            LOGGER.debug("Initiating shutdown of admin server.")
            shutdown.run(self.admin_server.stop())
        if self.inbound_transport_manager:
            LOGGER.debug("Initiating shutdown of inbound transport manager.")
            shutdown.run(self.inbound_transport_manager.stop())
        if self.outbound_transport_manager:
            LOGGER.debug("Initiating shutdown of outbound transport manager.")
            shutdown.run(self.outbound_transport_manager.stop())

        if self.root_profile:
            # close multitenant profiles
            multitenant_mgr = self.context.inject_or(BaseMultitenantManager)
            if multitenant_mgr:
                LOGGER.debug("Closing multitenant profiles.")
                for profile in multitenant_mgr.open_profiles:
                    LOGGER.debug("Closing profile: %s", profile.name)
                    shutdown.run(profile.close())
            LOGGER.debug("Closing root profile.")
            shutdown.run(self.root_profile.close())

        LOGGER.debug("Waiting for shutdown tasks to complete with timeout=%f.", timeout)
        await shutdown.complete(timeout)
        LOGGER.info("Conductor agent stopped successfully.")

    def inbound_message_router(
        self,
        profile: Profile,
        message: InboundMessage,
        can_respond: bool = False,
    ):
        """Route inbound messages.

        Args:
            profile: The active profile for the request
            message: The inbound message instance
            can_respond: If the session supports return routing

        """

        if message.receipt.direct_response_requested and not can_respond:
            LOGGER.warning(
                "Direct response requested, but not supported by transport: %s",
                message.transport_type,
            )

        # Note: at this point we could send the message to a shared queue
        # if this pod is too busy to process it

        try:
            self.dispatcher.queue_message(
                profile,
                message,
                self.outbound_message_router,
                lambda completed: self.dispatch_complete(message, completed),
            )
        except (LedgerConfigError, LedgerTransactionError) as e:
            LOGGER.error("Ledger error occurred in message handler: %s", str(e))
            raise

    def dispatch_complete(self, message: InboundMessage, completed: CompletedTask):
        """Handle completion of message dispatch."""
        if completed.exc_info:
            exc_class, exc, _ = completed.exc_info
            if isinstance(exc, (LedgerConfigError, LedgerTransactionError)):
                LOGGER.error(
                    "Ledger error occurred in message handler: %s",
                    str(exc),
                    exc_info=completed.exc_info,
                )
            elif isinstance(exc, (ProfileError, StorageNotFoundError)):
                LOGGER.error(
                    "Storage error occurred in message handler: %s: %s",
                    exc_class.__name__,
                    str(exc),
                    exc_info=completed.exc_info,
                )
            else:
                LOGGER.exception(
                    "Exception in message handler:", exc_info=completed.exc_info
                )

        self.inbound_transport_manager.dispatch_complete(message, completed)

    async def get_stats(self) -> dict:
        """Get the current stats tracked by the conductor."""
        stats = {
            "in_sessions": (
                len(self.inbound_transport_manager.sessions)
                if self.inbound_transport_manager
                else 0
            ),
            "out_encode": 0,
            "out_deliver": 0,
            "task_active": self.dispatcher.task_queue.current_active,
            "task_done": self.dispatcher.task_queue.total_done,
            "task_failed": self.dispatcher.task_queue.total_failed,
            "task_pending": self.dispatcher.task_queue.current_pending,
        }
        if self.outbound_transport_manager:
            for m in self.outbound_transport_manager.outbound_buffer:
                if m.state == QueuedOutboundMessage.STATE_ENCODE:
                    stats["out_encode"] += 1
                if m.state == QueuedOutboundMessage.STATE_DELIVER:
                    stats["out_deliver"] += 1
        return stats

    async def outbound_message_router(
        self,
        profile: Profile,
        outbound: OutboundMessage,
        inbound: Optional[InboundMessage] = None,
    ) -> OutboundSendStatus:
        """Route an outbound message.

        Args:
            profile: The active profile for the request
            outbound: An outbound message to be sent
            inbound: The inbound message that produced this response, if available
        """
        status: OutboundSendStatus = await self._outbound_message_router(
            profile=profile, outbound=outbound, inbound=inbound
        )
        await profile.notify(status.topic, outbound)
        return status

    async def _outbound_message_router(
        self,
        profile: Profile,
        outbound: OutboundMessage,
        inbound: Optional[InboundMessage] = None,
    ) -> OutboundSendStatus:
        """Route an outbound message.

        Args:
            profile: The active profile for the request
            outbound: An outbound message to be sent
            inbound: The inbound message that produced this response, if available
        """
        if not outbound.target and outbound.reply_to_verkey:
            if not outbound.reply_from_verkey and inbound:
                outbound.reply_from_verkey = inbound.receipt.recipient_verkey
            # return message to an inbound session
            if self.inbound_transport_manager.return_to_session(outbound):
                return OutboundSendStatus.SENT_TO_SESSION

        if not outbound.to_session_only:
            return await self.queue_outbound(profile, outbound, inbound)

    def handle_not_returned(self, profile: Profile, outbound: OutboundMessage):
        """Handle a message that failed delivery via an inbound session."""
        try:
            self.dispatcher.run_task(self.queue_outbound(profile, outbound))
        except (LedgerConfigError, LedgerTransactionError) as e:
            LOGGER.error(
                "Ledger error occurred while handling failed delivery: %s", str(e)
            )
            raise

    async def queue_outbound(
        self,
        profile: Profile,
        outbound: OutboundMessage,
        inbound: Optional[InboundMessage] = None,
    ) -> OutboundSendStatus:
        """Queue an outbound message for transport.

        Args:
            profile: The active profile
            outbound: The outbound message to be sent
            inbound: The inbound message that produced this response, if available
        """
        has_target = outbound.target or outbound.target_list

        # populate connection target(s)
        if not has_target and outbound.connection_id:
            conn_mgr = profile.inject(BaseConnectionManager)
            try:
                assert self.dispatcher, "dispatcher is not set"
                outbound.target_list = await self.dispatcher.run_task(
                    conn_mgr.get_connection_targets(connection_id=outbound.connection_id)
                )
            except BaseConnectionManagerError:
                LOGGER.exception("Error preparing outbound message for transmission")
                return OutboundSendStatus.UNDELIVERABLE
            except (LedgerConfigError, LedgerTransactionError) as e:
                LOGGER.error(
                    "Ledger error occurred while preparing outbound message: %s", str(e)
                )
                raise
            del conn_mgr
        # Find oob/connectionless target we can send the message to
        elif not has_target and outbound.reply_thread_id:
            message_processor = profile.inject(OobMessageProcessor)
            outbound.target = await self.dispatcher.run_task(
                message_processor.find_oob_target_for_outbound_message(profile, outbound)
            )

        return await self._queue_message(profile, outbound)

    async def _queue_message(
        self, profile: Profile, outbound: OutboundMessage
    ) -> OutboundSendStatus:
        """Save the message to an internal outbound queue."""
        try:
            await self.outbound_transport_manager.enqueue_message(profile, outbound)
            return OutboundSendStatus.QUEUED_FOR_DELIVERY
        except OutboundDeliveryError:
            LOGGER.warning("Cannot queue message for delivery, no supported transport")
            return self.handle_not_delivered(profile, outbound)

    def handle_not_delivered(
        self, profile: Profile, outbound: OutboundMessage
    ) -> OutboundSendStatus:
        """Handle a message that failed delivery via outbound transports."""
        queued_for_inbound = self.inbound_transport_manager.return_undelivered(outbound)
        return (
            OutboundSendStatus.WAITING_FOR_PICKUP
            if queued_for_inbound
            else OutboundSendStatus.UNDELIVERABLE
        )

    def webhook_router(
        self,
        topic: str,
        payload: dict,
        endpoint: str,
        max_attempts: Optional[int] = None,
        metadata: Optional[dict] = None,
    ):
        """Route a webhook through the outbound transport manager.

        Args:
            topic: The webhook topic
            payload: The webhook payload
            endpoint: The endpoint of the webhook target
            max_attempts: The maximum number of attempts
            metadata: Additional metadata associated with the payload
        """
        try:
            self.outbound_transport_manager.enqueue_webhook(
                topic, payload, endpoint, max_attempts, metadata
            )
        except OutboundDeliveryError:
            LOGGER.warning(
                "Cannot queue message webhook for delivery, no supported transport"
            )

    async def check_for_valid_wallet_type(self, profile):
        """Check wallet type and set it if not set. Raise an error if wallet type config doesn't match existing storage type."""  # noqa: E501
        async with profile.session() as session:
            storage_type_from_config = profile.settings.get("wallet.type")
            storage = session.inject(BaseStorage)
            try:
                storage_type_record = await storage.find_record(
                    type_filter=RECORD_TYPE_ACAPY_STORAGE_TYPE, tag_query={}
                )
                storage_type_from_storage = storage_type_record.value
            except StorageNotFoundError:
                storage_type_record = None

            if not storage_type_record:
                LOGGER.info("Wallet type record not found.")
                try:
                    acapy_version = await storage.find_record(
                        type_filter=RECORD_TYPE_ACAPY_VERSION, tag_query={}
                    )
                except StorageNotFoundError:
                    acapy_version = None
                # Any existing agent will have acapy_version record
                if acapy_version:
                    storage_type_from_storage = STORAGE_TYPE_VALUE_ASKAR
                    LOGGER.info(
                        "Existing agent found. Setting wallet type to %s.",
                        storage_type_from_storage,
                    )
                    await storage.add_record(
                        StorageRecord(
                            RECORD_TYPE_ACAPY_STORAGE_TYPE,
                            storage_type_from_storage,
                        )
                    )
                else:
                    storage_type_from_storage = storage_type_from_config
                    LOGGER.info(
                        "New agent. Setting wallet type to %s.", storage_type_from_config
                    )
                    await storage.add_record(
                        StorageRecord(
                            RECORD_TYPE_ACAPY_STORAGE_TYPE,
                            storage_type_from_config,
                        )
                    )

            if storage_type_from_storage != storage_type_from_config:
                if (
                    storage_type_from_config == STORAGE_TYPE_VALUE_ASKAR
                    and storage_type_from_storage == STORAGE_TYPE_VALUE_ANONCREDS
                ):
                    LOGGER.warning(
                        "The agent has been upgrade to use anoncreds wallet. Please update the wallet.type in the config file to 'askar-anoncreds'"  # noqa: E501
                    )
                    # Allow agent to create anoncreds profile with askar
                    # wallet type config by stopping conductor and reloading context
                    await self.stop()
                    self.force_agent_anoncreds = True
                    self.context.settings.set_value("wallet.type", "askar-anoncreds")
                    self.context_builder = DefaultContextBuilder(self.context.settings)
                    await self.setup()
                else:
                    raise StartupError(
                        f"Wallet type config [{storage_type_from_config}] doesn't match with the wallet type in storage [{storage_type_record.value}]"  # noqa: E501
                    )

    async def check_for_wallet_upgrades_in_progress(self):
        """Check for upgrade and upgrade if needed."""
        if self.context.settings.get_value("multitenant.enabled"):
            # Sub-wallets
            subwallet_profiles = await get_subwallet_profiles_from_storage(
                self.root_profile
            )
            await asyncio.gather(
                *[
                    upgrade_wallet_to_anoncreds_if_requested(profile, is_subwallet=True)
                    for profile in subwallet_profiles
                ]
            )

        # Stand-alone or admin wallet
        await upgrade_wallet_to_anoncreds_if_requested(self.root_profile)



"""Test Revoke Message."""

from ..revoke import Revoke


def test_instantiate():
    msg = Revoke(
        revocation_format="indy-anoncreds",
        credential_id="test-id",
        comment="test",
    )
    assert msg.revocation_format == "indy-anoncreds"
    assert msg.credential_id == "test-id"
    assert msg.comment == "test"



"""A message responder.

The responder is provided to message handlers to enable them to send a new message
in response to the message being handled.
"""

import asyncio
import json
from abc import ABC, abstractmethod
from typing import List, Optional, Sequence, Tuple, Union

from ..cache.base import BaseCache
from ..connections.models.conn_record import ConnRecord
from ..connections.models.connection_target import ConnectionTarget
from ..core.error import BaseError
from ..core.profile import Profile
from ..transport.outbound.message import OutboundMessage
from ..transport.outbound.status import OutboundSendStatus
from .base_message import BaseMessage

SKIP_ACTIVE_CONN_CHECK_MSG_TYPES = [
    "didexchange/1.1/request",
    "didexchange/1.1/response",
    "didexchange/1.1/problem_report",
    "didexchange/1.0/request",
    "didexchange/1.0/response",
    "didexchange/1.0/problem_report",
    "connections/1.0/invitation",
    "connections/1.0/request",
    "connections/1.0/response",
    "connections/1.0/problem_report",
]


class ResponderError(BaseError):
    """Responder error."""


class BaseResponder(ABC):
    """Interface for message handlers to send responses."""

    def __init__(
        self,
        *,
        connection_id: Optional[str] = None,
        reply_session_id: Optional[str] = None,
        reply_to_verkey: Optional[str] = None,
    ):
        """Initialize a base responder."""
        self.connection_id = connection_id
        self.reply_session_id = reply_session_id
        self.reply_to_verkey = reply_to_verkey

    async def create_outbound(
        self,
        message: Union[BaseMessage, str, bytes],
        *,
        connection_id: Optional[str] = None,
        reply_session_id: Optional[str] = None,
        reply_thread_id: Optional[str] = None,
        reply_to_verkey: Optional[str] = None,
        reply_from_verkey: Optional[str] = None,
        target: Optional[ConnectionTarget] = None,
        target_list: Sequence[ConnectionTarget] = None,
        to_session_only: bool = False,
    ) -> OutboundMessage:
        """Create an OutboundMessage from a message payload."""
        if isinstance(message, BaseMessage):
            # TODO DIDComm version selection
            serialized = message.serialize()
            # TODO serialized format selection?
            payload = json.dumps(serialized)
            enc_payload = None
            if not reply_thread_id:
                reply_thread_id = message._thread_id
        else:
            payload = None
            enc_payload = message
        return OutboundMessage(
            connection_id=connection_id,
            enc_payload=enc_payload,
            payload=payload,
            reply_session_id=reply_session_id,
            reply_thread_id=reply_thread_id,
            reply_to_verkey=reply_to_verkey,
            reply_from_verkey=reply_from_verkey,
            target=target,
            target_list=target_list,
            to_session_only=to_session_only,
        )

    async def send(
        self, message: Union[BaseMessage, str, bytes], **kwargs
    ) -> OutboundSendStatus:
        """Convert a message to an OutboundMessage and send it."""
        outbound = await self.create_outbound(message, **kwargs)
        if isinstance(message, BaseMessage):
            msg_type = message._message_type
            msg_id = message._id
        else:
            msg_dict = json.loads(message)
            msg_type = msg_dict.get("@type")
            msg_id = msg_dict.get("@id")
        return await self.send_outbound(
            message=outbound,
            message_type=msg_type,
            message_id=msg_id,
        )

    async def send_reply(
        self,
        message: Union[BaseMessage, str, bytes],
        *,
        connection_id: Optional[str] = None,
        target: Optional[ConnectionTarget] = None,
        target_list: Sequence[ConnectionTarget] = None,
    ) -> OutboundSendStatus:
        """Send a reply to an incoming message.

        Args:
            message: the `BaseMessage`, or pre-packed str or bytes to reply with
            connection_id: optionally override the target connection ID
            target: optionally specify a `ConnectionTarget` to send to
            target_list: optionally specify a list of `ConnectionTarget` to send to

        Raises:
            ResponderError: If there is no active connection

        """
        outbound = await self.create_outbound(
            message,
            connection_id=connection_id or self.connection_id,
            reply_session_id=self.reply_session_id,
            reply_to_verkey=self.reply_to_verkey,
            target=target,
            target_list=target_list,
        )
        if isinstance(message, BaseMessage):
            msg_type = message._message_type
            msg_id = message._id
        else:
            msg_dict = json.loads(message)
            msg_type = msg_dict.get("@type")
            msg_id = msg_dict.get("@id")
        return await self.send_outbound(
            message=outbound, message_type=msg_type, message_id=msg_id
        )

    async def conn_rec_active_state_check(
        self, profile: Profile, connection_id: str, timeout: int = 7
    ) -> bool:
        """Check if the connection record is ready for sending outbound message."""

        async def _wait_for_state() -> Tuple[bool, Optional[str]]:
            while True:
                async with profile.session() as session:
                    conn_record = await ConnRecord.retrieve_by_id(session, connection_id)
                    if conn_record.is_ready:
                        # if ConnRecord.State.get(conn_record.state) in (
                        #     ConnRecord.State.COMPLETED,
                        # ):
                        return (True, conn_record.state)
                    await asyncio.sleep(1)

        try:
            cache_key = f"conn_rec_state::{connection_id}"
            connection_state = None
            cache = profile.inject_or(BaseCache)
            if cache:
                connection_state = await cache.get(cache_key)
            if connection_state and ConnRecord.State.get(connection_state) in (
                ConnRecord.State.COMPLETED,
                ConnRecord.State.RESPONSE,
            ):
                return True
            check_flag, connection_state = await asyncio.wait_for(
                _wait_for_state(), timeout
            )
            if cache and connection_state:
                await cache.set(cache_key, connection_state)
            return check_flag
        except asyncio.TimeoutError:
            return False

    @abstractmethod
    async def send_outbound(
        self, message: OutboundMessage, **kwargs
    ) -> OutboundSendStatus:
        """Send an outbound message.

        Args:
            message: The `OutboundMessage` to be sent
            kwargs: Additional keyword arguments
        """

    @abstractmethod
    async def send_webhook(self, topic: str, payload: dict):
        """Dispatch a webhook. DEPRECATED: use the event bus instead.

        Args:
            topic: the webhook topic identifier
            payload: the webhook payload value
        """


class MockResponder(BaseResponder):
    """Mock responder implementation for use by tests."""

    def __init__(self):
        """Initialize the mock responder."""
        self.messages: List[
            Tuple[Union[BaseMessage, str, bytes, OutboundMessage], Optional[dict]]
        ] = []

    async def send(
        self, message: Union[BaseMessage, str, bytes], **kwargs
    ) -> OutboundSendStatus:
        """Convert a message to an OutboundMessage and send it."""
        self.messages.append((message, kwargs))
        return OutboundSendStatus.QUEUED_FOR_DELIVERY

    async def send_reply(
        self, message: Union[BaseMessage, str, bytes], **kwargs
    ) -> OutboundSendStatus:
        """Send a reply to an incoming message."""
        self.messages.append((message, kwargs))
        return OutboundSendStatus.QUEUED_FOR_DELIVERY

    async def send_outbound(
        self, message: OutboundMessage, **kwargs
    ) -> OutboundSendStatus:
        """Send an outbound message."""
        self.messages.append((message, None))
        return OutboundSendStatus.QUEUED_FOR_DELIVERY

    async def send_webhook(self, topic: str, payload: dict):
        """Send an outbound message."""
        raise Exception(
            "responder.send_webhook is deprecated; please use the event bus instead."
        )



"""Ping response handler."""

from .....messaging.base_handler import BaseHandler, BaseResponder, RequestContext
from ..messages.ping_response import PingResponse


class PingResponseHandler(BaseHandler):
    """Ping response handler class."""

    async def handle(self, context: RequestContext, responder: BaseResponder):
        """Handle ping response message.

        Args:
            context: Request context
            responder: Responder used to reply

        """

        self._logger.debug("PingResponseHandler called with context: %s", context)
        assert isinstance(context.message, PingResponse)

        self._logger.debug(
            "Received trust ping response from: %s", context.message_receipt.sender_did
        )

        if context.settings.get("debug.monitor_ping"):
            await context.profile.notify(
                "acapy::ping::response_received",
                {
                    "comment": context.message.comment,
                    "connection_id": context.message_receipt.connection_id,
                    "state": "response_received",
                    "thread_id": context.message._thread_id,
                },
            )

        # Nothing to do, Connection should be automatically promoted to 'active'



from unittest import TestCase

import pytest

from ......messaging.decorators.attach_decorator import AttachDecorator
from ......messaging.models.base import BaseModelError
from .....didcomm_prefix import DIDCommPrefix
from ...message_types import ATTACHMENT_FORMAT, PRES_20_PROPOSAL
from ..pres_format import V20PresFormat
from ..pres_proposal import V20PresProposal

CD_ID = "NcYxiDXkpYi6ov5FcYDi1e:3:CL:12:tag1"
INDY_PROOF_REQ = [
    {
        "name": "proof-req",
        "version": "1.0",
        "nonce": "12345",
        "requested_attributes": {
            "0_player_uuid": {
                "name": "player",
                "restrictions": [
                    {
                        "cred_def_id": f"{CD_ID}",
                        "attr::player::value": "Richie Knucklez",
                    }
                ],
                "non_revoked": {
                    "from": 1234567890,
                    "to": 1234567890,
                },
            },
            "0_screencapture_uuid": {
                "name": "screenCapture",
                "restrictions": [{"cred_def_id": f"{CD_ID}"}],
                "non_revoked": {
                    "from": 1234567890,
                    "to": 1234567890,
                },
            },
        },
        "requested_predicates": {
            "0_highscore_GE_uuid": {
                "name": "highScore",
                "p_type": ">=",
                "p_value": 1000000,
                "restrictions": [{"cred_def_id": f"{CD_ID}"}],
                "non_revoked": {
                    "from": 1234567890,
                    "to": 1234567890,
                },
            }
        },
    },
    {
        "name": "proof-req",
        "version": "1.0",
        "nonce": "123456",
        "requested_attributes": {
            "0_player_uuid": {
                "name": "player",
                "restrictions": [{"cred_def_id": f"{CD_ID}"}],
            },
            "0_screencapture_uuid": {
                "name": "screenCapture",
                "restrictions": [{"cred_def_id": f"{CD_ID}"}],
            },
        },
        "requested_predicates": {
            "0_highscore_GE_uuid": {
                "name": "highScore",
                "p_type": ">=",
                "p_value": 1000000,
                "restrictions": [{"cred_def_id": f"{CD_ID}"}],
            }
        },
    },
    {
        "name": "proof-req",
        "version": "1.0",
        "nonce": "1234567",
        "requested_attributes": {},
        "requested_predicates": {
            "0_highscore_GE_uuid": {
                "name": "highScore",
                "p_type": ">=",
                "p_value": 1000000,
                "restrictions": [{"cred_def_id": f"{CD_ID}"}],
            }
        },
    },
]


class TestV20PresProposal(TestCase):
    """Presentation proposal tests."""

    def test_init_type_attachment_serde(self):
        """Test initializer, type, attachment, de/serialization."""
        for proof_req in INDY_PROOF_REQ:
            pres_proposal = V20PresProposal(
                comment="Hello World",
                formats=[
                    V20PresFormat(
                        attach_id="indy",
                        format_=ATTACHMENT_FORMAT[PRES_20_PROPOSAL][
                            V20PresFormat.Format.INDY.api
                        ],
                    )
                ],
                proposals_attach=[AttachDecorator.data_base64(proof_req, ident="indy")],
            )
            assert pres_proposal._type == DIDCommPrefix.qualify_current(PRES_20_PROPOSAL)
            assert pres_proposal.attachment(V20PresFormat.Format.INDY) == proof_req

            pres_proposal_ser = pres_proposal.serialize()
            pres_proposal_deser = V20PresProposal.deserialize(pres_proposal_ser)
            assert type(pres_proposal_deser) is V20PresProposal

            pres_proposal_dict = pres_proposal_deser.serialize()
            assert pres_proposal_dict == pres_proposal_ser

            pres_proposal_dict["formats"][0]["attach_id"] = "xxx"
            with pytest.raises(BaseModelError):
                V20PresProposal.deserialize(pres_proposal_dict)  # id mismatch

            pres_proposal_dict["formats"] = []
            with pytest.raises(BaseModelError):
                V20PresProposal.deserialize(pres_proposal_dict)  # length mismatch

            pres_proposal.formats.append(  # unknown format: no validation
                V20PresFormat(
                    attach_id="not_indy",
                    format_="not_indy",
                )
            )
            obj = pres_proposal.serialize()
            obj["proposals~attach"].append(
                {
                    "@id": "not_indy",
                    "mime-type": "application/json",
                    "data": {"base64": "eyJub3QiOiAiaW5keSJ9"},
                }
            )
            V20PresProposal.deserialize(obj)

    def test_attachment_no_target_format(self):
        """Test attachment behaviour for only unknown formats."""

        x_pres_proposal = V20PresProposal(
            comment="Test",
            formats=[V20PresFormat(attach_id="not_indy", format_="not_indy")],
            proposals_attach=[
                AttachDecorator.data_base64(ident="not_indy", mapping=INDY_PROOF_REQ[0])
            ],
        )
        assert x_pres_proposal.attachment() is None



"""Credential exchange admin routes."""

from json.decoder import JSONDecodeError
from typing import Optional

from aiohttp import web
from aiohttp_apispec import (
    docs,
    match_info_schema,
    querystring_schema,
    request_schema,
    response_schema,
)
from marshmallow import fields, validate

from ....admin.decorators.auth import tenant_authentication
from ....admin.request_context import AdminRequestContext
from ....connections.models.conn_record import ConnRecord
from ....core.profile import Profile
from ....indy.holder import IndyHolderError
from ....indy.issuer import IndyIssuerError
from ....ledger.error import LedgerError
from ....messaging.credential_definitions.util import CRED_DEF_TAGS
from ....messaging.models.base import BaseModelError
from ....messaging.models.openapi import OpenAPISchema
from ....messaging.models.paginated_query import (
    PaginatedQuerySchema,
    get_paginated_query_params,
)
from ....messaging.valid import (
    INDY_CRED_DEF_ID_EXAMPLE,
    INDY_CRED_DEF_ID_VALIDATE,
    INDY_DID_EXAMPLE,
    INDY_DID_VALIDATE,
    INDY_SCHEMA_ID_EXAMPLE,
    INDY_SCHEMA_ID_VALIDATE,
    MAJOR_MINOR_VERSION_EXAMPLE,
    MAJOR_MINOR_VERSION_VALIDATE,
    UUID4_EXAMPLE,
    UUID4_VALIDATE,
)
from ....storage.error import StorageError, StorageNotFoundError
from ....utils.tracing import AdminAPIMessageTracingSchema, get_timer, trace_event
from ....wallet.util import default_did_from_verkey
from ...out_of_band.v1_0.models.oob_record import OobRecord
from . import problem_report_for_record, report_problem
from .manager import CredentialManager, CredentialManagerError
from .message_types import SPEC_URI
from .messages.credential_problem_report import ProblemReportReason
from .messages.credential_proposal import CredentialProposal, CredentialProposalSchema
from .messages.inner.credential_preview import CredentialPreview, CredentialPreviewSchema
from .models.credential_exchange import V10CredentialExchange, V10CredentialExchangeSchema


class IssueCredentialModuleResponseSchema(OpenAPISchema):
    """Response schema for Issue Credential Module."""


class V10CredentialExchangeListQueryStringSchema(PaginatedQuerySchema):
    """Parameters and validators for credential exchange list query."""

    connection_id = fields.Str(
        required=False,
        metadata={"description": "Connection identifier", "example": UUID4_EXAMPLE},
    )
    thread_id = fields.Str(
        required=False,
        metadata={"description": "Thread identifier", "example": UUID4_EXAMPLE},
    )
    role = fields.Str(
        required=False,
        validate=validate.OneOf(
            [
                getattr(V10CredentialExchange, m)
                for m in vars(V10CredentialExchange)
                if m.startswith("ROLE_")
            ]
        ),
        metadata={"description": "Role assigned in credential exchange"},
    )
    state = fields.Str(
        required=False,
        validate=validate.OneOf(
            [
                getattr(V10CredentialExchange, m)
                for m in vars(V10CredentialExchange)
                if m.startswith("STATE_")
            ]
        ),
        metadata={"description": "Credential exchange state"},
    )


class V10CredentialExchangeListResultSchema(OpenAPISchema):
    """Result schema for Aries#0036 v1.0 credential exchange query."""

    results = fields.List(
        fields.Nested(V10CredentialExchangeSchema),
        metadata={"description": "Aries#0036 v1.0 credential exchange records"},
    )


class V10CredentialStoreRequestSchema(OpenAPISchema):
    """Request schema for sending a credential store admin message."""

    credential_id = fields.Str(required=False)


class V10CredentialCreateSchema(AdminAPIMessageTracingSchema):
    """Base class for request schema for sending credential proposal admin message."""

    cred_def_id = fields.Str(
        required=False,
        validate=INDY_CRED_DEF_ID_VALIDATE,
        metadata={
            "description": "Credential definition identifier",
            "example": INDY_CRED_DEF_ID_EXAMPLE,
        },
    )
    schema_id = fields.Str(
        required=False,
        validate=INDY_SCHEMA_ID_VALIDATE,
        metadata={
            "description": "Schema identifier",
            "example": INDY_SCHEMA_ID_EXAMPLE,
        },
    )
    schema_issuer_did = fields.Str(
        required=False,
        validate=INDY_DID_VALIDATE,
        metadata={"description": "Schema issuer DID", "example": INDY_DID_EXAMPLE},
    )
    schema_name = fields.Str(
        required=False,
        metadata={"description": "Schema name", "example": "preferences"},
    )
    schema_version = fields.Str(
        required=False,
        validate=MAJOR_MINOR_VERSION_VALIDATE,
        metadata={
            "description": "Schema version",
            "example": MAJOR_MINOR_VERSION_EXAMPLE,
        },
    )
    issuer_did = fields.Str(
        required=False,
        validate=INDY_DID_VALIDATE,
        metadata={"description": "Credential issuer DID", "example": INDY_DID_EXAMPLE},
    )
    auto_remove = fields.Bool(
        required=False,
        metadata={
            "description": (
                "Whether to remove the credential exchange record on completion"
                " (overrides --preserve-exchange-records configuration setting)"
            )
        },
    )
    comment = fields.Str(
        required=False,
        allow_none=True,
        metadata={"description": "Human-readable comment"},
    )
    credential_proposal = fields.Nested(CredentialPreviewSchema, required=True)


class V10CredentialProposalRequestSchemaBase(AdminAPIMessageTracingSchema):
    """Base class for request schema for sending credential proposal admin message."""

    connection_id = fields.Str(
        required=True,
        metadata={"description": "Connection identifier", "example": UUID4_EXAMPLE},
    )
    cred_def_id = fields.Str(
        required=False,
        validate=INDY_CRED_DEF_ID_VALIDATE,
        metadata={
            "description": "Credential definition identifier",
            "example": INDY_CRED_DEF_ID_EXAMPLE,
        },
    )
    schema_id = fields.Str(
        required=False,
        validate=INDY_SCHEMA_ID_VALIDATE,
        metadata={
            "description": "Schema identifier",
            "example": INDY_SCHEMA_ID_EXAMPLE,
        },
    )
    schema_issuer_did = fields.Str(
        required=False,
        validate=INDY_DID_VALIDATE,
        metadata={"description": "Schema issuer DID", "example": INDY_DID_EXAMPLE},
    )
    schema_name = fields.Str(
        required=False,
        metadata={"description": "Schema name", "example": "preferences"},
    )
    schema_version = fields.Str(
        required=False,
        validate=MAJOR_MINOR_VERSION_VALIDATE,
        metadata={
            "description": "Schema version",
            "example": MAJOR_MINOR_VERSION_EXAMPLE,
        },
    )
    issuer_did = fields.Str(
        required=False,
        validate=INDY_DID_VALIDATE,
        metadata={"description": "Credential issuer DID", "example": INDY_DID_EXAMPLE},
    )
    auto_remove = fields.Bool(
        required=False,
        metadata={
            "description": (
                "Whether to remove the credential exchange record on completion"
                " (overrides --preserve-exchange-records configuration setting)"
            )
        },
    )
    comment = fields.Str(
        required=False,
        allow_none=True,
        metadata={"description": "Human-readable comment"},
    )


class V10CredentialProposalRequestOptSchema(V10CredentialProposalRequestSchemaBase):
    """Request schema for sending credential proposal on optional proposal preview."""

    credential_proposal = fields.Nested(CredentialPreviewSchema, required=False)


class V10CredentialProposalRequestMandSchema(V10CredentialProposalRequestSchemaBase):
    """Request schema for sending credential proposal on mandatory proposal preview."""

    credential_proposal = fields.Nested(CredentialPreviewSchema, required=True)


class V10CredentialBoundOfferRequestSchema(OpenAPISchema):
    """Request schema for sending bound credential offer admin message."""

    counter_proposal = fields.Nested(
        CredentialProposalSchema,
        required=False,
        metadata={"description": "Optional counter-proposal"},
    )


class V10CredentialFreeOfferRequestSchema(AdminAPIMessageTracingSchema):
    """Request schema for sending free credential offer admin message."""

    connection_id = fields.Str(
        required=True,
        metadata={"description": "Connection identifier", "example": UUID4_EXAMPLE},
    )
    cred_def_id = fields.Str(
        required=True,
        validate=INDY_CRED_DEF_ID_VALIDATE,
        metadata={
            "description": "Credential definition identifier",
            "example": INDY_CRED_DEF_ID_EXAMPLE,
        },
    )
    auto_issue = fields.Bool(
        required=False,
        metadata={
            "description": (
                "Whether to respond automatically to credential requests, creating and"
                " issuing requested credentials"
            )
        },
    )
    auto_remove = fields.Bool(
        required=False,
        dump_default=True,
        metadata={
            "description": (
                "Whether to remove the credential exchange record on completion"
                " (overrides --preserve-exchange-records configuration setting)"
            )
        },
    )
    comment = fields.Str(
        required=False,
        allow_none=True,
        metadata={"description": "Human-readable comment"},
    )
    credential_preview = fields.Nested(CredentialPreviewSchema, required=True)


class V10CredentialConnFreeOfferRequestSchema(AdminAPIMessageTracingSchema):
    """Request schema for creating connection free credential offer."""

    cred_def_id = fields.Str(
        required=True,
        validate=INDY_CRED_DEF_ID_VALIDATE,
        metadata={
            "description": "Credential definition identifier",
            "example": INDY_CRED_DEF_ID_EXAMPLE,
        },
    )
    auto_issue = fields.Bool(
        required=False,
        metadata={
            "description": (
                "Whether to respond automatically to credential requests, creating and"
                " issuing requested credentials"
            )
        },
    )
    auto_remove = fields.Bool(
        required=False,
        dump_default=True,
        metadata={
            "description": (
                "Whether to remove the credential exchange record on completion"
                " (overrides --preserve-exchange-records configuration setting)"
            )
        },
    )
    comment = fields.Str(
        required=False,
        allow_none=True,
        metadata={"description": "Human-readable comment"},
    )
    credential_preview = fields.Nested(CredentialPreviewSchema, required=True)


class V10CredentialIssueRequestSchema(OpenAPISchema):
    """Request schema for sending credential issue admin message."""

    comment = fields.Str(
        required=False,
        allow_none=True,
        metadata={"description": "Human-readable comment"},
    )


class V10CredentialProblemReportRequestSchema(OpenAPISchema):
    """Request schema for sending problem report."""

    description = fields.Str(required=True)


class CredIdMatchInfoSchema(OpenAPISchema):
    """Path parameters and validators for request taking credential id."""

    credential_id = fields.Str(
        required=True,
        metadata={"description": "Credential identifier", "example": UUID4_EXAMPLE},
    )


class CredExIdMatchInfoSchema(OpenAPISchema):
    """Path parameters and validators for request taking credential exchange id."""

    cred_ex_id = fields.Str(
        required=True,
        validate=UUID4_VALIDATE,
        metadata={
            "description": "Credential exchange identifier",
            "example": UUID4_EXAMPLE,
        },
    )


class V10CredentialExchangeAutoRemoveRequestSchema(OpenAPISchema):
    """Request Schema for overriding default preserve exchange records setting."""

    auto_remove = fields.Bool(
        required=False,
        dump_default=False,
        metadata={
            "description": (
                "Whether to remove the credential exchange record on completion"
                " (overrides --preserve-exchange-records configuration setting)"
            )
        },
    )


@docs(
    tags=["issue-credential v1.0"],
    summary="Fetch all credential exchange records",
    deprecated=True,
)
@querystring_schema(V10CredentialExchangeListQueryStringSchema)
@response_schema(V10CredentialExchangeListResultSchema(), 200, description="")
@tenant_authentication
async def credential_exchange_list(request: web.BaseRequest):
    """Request handler for searching credential exchange records.

    Args:
        request: aiohttp request object

    Returns:
        The connection list response

    """
    context: AdminRequestContext = request["context"]
    tag_filter = {}
    if "thread_id" in request.query and request.query["thread_id"] != "":
        tag_filter["thread_id"] = request.query["thread_id"]
    post_filter = {
        k: request.query[k]
        for k in ("connection_id", "role", "state")
        if request.query.get(k, "") != ""
    }

    limit, offset, order_by, descending = get_paginated_query_params(request)

    try:
        async with context.profile.session() as session:
            records = await V10CredentialExchange.query(
                session=session,
                tag_filter=tag_filter,
                limit=limit,
                offset=offset,
                order_by=order_by,
                descending=descending,
                post_filter_positive=post_filter,
            )
        results = [record.serialize() for record in records]
    except (StorageError, BaseModelError) as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    return web.json_response({"results": results})


@docs(
    tags=["issue-credential v1.0"],
    summary="Fetch a single credential exchange record",
    deprecated=True,
)
@match_info_schema(CredExIdMatchInfoSchema())
@response_schema(V10CredentialExchangeSchema(), 200, description="")
@tenant_authentication
async def credential_exchange_retrieve(request: web.BaseRequest):
    """Request handler for fetching single credential exchange record.

    Args:
        request: aiohttp request object

    Returns:
        The credential exchange record

    """
    context: AdminRequestContext = request["context"]
    outbound_handler = request["outbound_message_router"]

    credential_exchange_id = request.match_info["cred_ex_id"]
    cred_ex_record = None
    try:
        async with context.profile.session() as session:
            cred_ex_record = await V10CredentialExchange.retrieve_by_id(
                session, credential_exchange_id
            )
        result = cred_ex_record.serialize()
    except StorageNotFoundError as err:
        # no such cred ex record: not protocol error, user fat-fingered id
        raise web.HTTPNotFound(reason=err.roll_up) from err
    except (BaseModelError, StorageError) as err:
        # present but broken or hopeless: protocol error
        await report_problem(
            err,
            ProblemReportReason.ISSUANCE_ABANDONED.value,
            web.HTTPBadRequest,
            cred_ex_record,
            outbound_handler,
        )

    return web.json_response(result)


@docs(
    tags=["issue-credential v1.0"],
    summary=(
        "Create a credential record without sending (generally for use with Out-Of-Band)"
    ),
    deprecated=True,
)
@request_schema(V10CredentialCreateSchema())
@response_schema(V10CredentialExchangeSchema(), 200, description="")
@tenant_authentication
async def credential_exchange_create(request: web.BaseRequest):
    """Request handler for creating a credential from attr values.

    The internal credential record will be created without the credential
    being sent to any connection. This can be used in conjunction with
    the `oob` protocols to bind messages to an out of band message.

    Args:
        request: aiohttp request object

    Returns:
        The credential exchange record

    """
    r_time = get_timer()

    context: AdminRequestContext = request["context"]
    profile = context.profile
    body = await request.json()

    comment = body.get("comment")
    preview_spec = body.get("credential_proposal")
    if not preview_spec:
        raise web.HTTPBadRequest(reason="credential_proposal must be provided")
    auto_remove = body.get(
        "auto_remove", not profile.settings.get("preserve_exchange_records")
    )
    trace_msg = body.get("trace")

    try:
        preview = CredentialPreview.deserialize(preview_spec)

        credential_proposal = CredentialProposal(
            comment=comment,
            credential_proposal=preview,
            **{t: body.get(t) for t in CRED_DEF_TAGS if body.get(t)},
        )
        credential_proposal.assign_trace_decorator(
            context.settings,
            trace_msg,
        )

        trace_event(
            context.settings,
            credential_proposal,
            outcome="credential_exchange_create.START",
        )

        credential_manager = CredentialManager(context.profile)

        (
            credential_exchange_record,
            credential_offer_message,
        ) = await credential_manager.prepare_send(
            None,
            credential_proposal=credential_proposal,
            auto_remove=auto_remove,
            comment=comment,
        )
    except (StorageError, BaseModelError) as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    trace_event(
        context.settings,
        credential_offer_message,
        outcome="credential_exchange_create.END",
        perf_counter=r_time,
    )

    return web.json_response(credential_exchange_record.serialize())


@docs(
    tags=["issue-credential v1.0"],
    summary="Send holder a credential, automating entire flow",
    deprecated=True,
)
@request_schema(V10CredentialProposalRequestMandSchema())
@response_schema(V10CredentialExchangeSchema(), 200, description="")
@tenant_authentication
async def credential_exchange_send(request: web.BaseRequest):
    """Request handler for sending credential from issuer to holder from attr values.

    If both issuer and holder are configured for automatic responses, the operation
    ultimately results in credential issue; otherwise, the result waits on the first
    response not automated; the credential exchange record retains state regardless.

    Args:
        request: aiohttp request object

    Returns:
        The credential exchange record

    """
    r_time = get_timer()

    context: AdminRequestContext = request["context"]
    profile = context.profile
    outbound_handler = request["outbound_message_router"]

    body = await request.json()

    comment = body.get("comment")
    connection_id = body.get("connection_id")
    preview_spec = body.get("credential_proposal")
    if not preview_spec:
        raise web.HTTPBadRequest(reason="credential_proposal must be provided")
    auto_remove = body.get(
        "auto_remove", not profile.settings.get("preserve_exchange_records")
    )
    trace_msg = body.get("trace")

    connection_record = None
    cred_ex_record = None
    try:
        preview = CredentialPreview.deserialize(preview_spec)
        async with profile.session() as session:
            connection_record = await ConnRecord.retrieve_by_id(session, connection_id)
            if not connection_record.is_ready:
                raise web.HTTPForbidden(reason=f"Connection {connection_id} not ready")

        credential_proposal = CredentialProposal(
            comment=comment,
            credential_proposal=preview,
            **{t: body.get(t) for t in CRED_DEF_TAGS if body.get(t)},
        )
        credential_proposal.assign_trace_decorator(
            context.settings,
            trace_msg,
        )

        trace_event(
            context.settings,
            credential_proposal,
            outcome="credential_exchange_send.START",
        )

        credential_manager = CredentialManager(profile)
        (
            cred_ex_record,
            credential_offer_message,
        ) = await credential_manager.prepare_send(
            connection_id,
            credential_proposal=credential_proposal,
            auto_remove=auto_remove,
            comment=comment,
        )
        result = cred_ex_record.serialize()

    except (BaseModelError, CredentialManagerError, LedgerError, StorageError) as err:
        if cred_ex_record:
            async with profile.session() as session:
                await cred_ex_record.save_error_state(session, reason=err.roll_up)
        await report_problem(
            err,
            ProblemReportReason.ISSUANCE_ABANDONED.value,
            web.HTTPBadRequest,
            cred_ex_record or connection_record,
            outbound_handler,
        )

    await outbound_handler(
        credential_offer_message, connection_id=cred_ex_record.connection_id
    )

    trace_event(
        context.settings,
        credential_offer_message,
        outcome="credential_exchange_send.END",
        perf_counter=r_time,
    )

    return web.json_response(result)


@docs(
    tags=["issue-credential v1.0"],
    summary="Send issuer a credential proposal",
    deprecated=True,
)
@request_schema(V10CredentialProposalRequestOptSchema())
@response_schema(V10CredentialExchangeSchema(), 200, description="")
@tenant_authentication
async def credential_exchange_send_proposal(request: web.BaseRequest):
    """Request handler for sending credential proposal.

    Args:
        request: aiohttp request object

    Returns:
        The credential exchange record

    """
    r_time = get_timer()

    context: AdminRequestContext = request["context"]
    profile = context.profile
    outbound_handler = request["outbound_message_router"]

    body = await request.json()

    connection_id = body.get("connection_id")
    comment = body.get("comment")
    preview_spec = body.get("credential_proposal")
    auto_remove = body.get(
        "auto_remove", not profile.settings.get("preserve_exchange_records")
    )
    trace_msg = body.get("trace")

    connection_record = None
    cred_ex_record = None
    try:
        preview = CredentialPreview.deserialize(preview_spec) if preview_spec else None
        async with profile.session() as session:
            connection_record = await ConnRecord.retrieve_by_id(session, connection_id)
            if not connection_record.is_ready:
                raise web.HTTPForbidden(reason=f"Connection {connection_id} not ready")

        credential_manager = CredentialManager(profile)
        cred_ex_record = await credential_manager.create_proposal(
            connection_id,
            comment=comment,
            credential_preview=preview,
            auto_remove=auto_remove,
            trace=trace_msg,
            **{t: body.get(t) for t in CRED_DEF_TAGS if body.get(t)},
        )

        credential_proposal = cred_ex_record.credential_proposal_dict
        result = cred_ex_record.serialize()

    except (BaseModelError, StorageError) as err:
        if cred_ex_record:
            async with profile.session() as session:
                await cred_ex_record.save_error_state(session, reason=err.roll_up)
        # other party cannot yet receive a problem report about our failed protocol start
        raise web.HTTPBadRequest(reason=err.roll_up)

    await outbound_handler(
        credential_proposal,
        connection_id=connection_id,
    )

    trace_event(
        context.settings,
        credential_proposal,
        outcome="credential_exchange_send_proposal.END",
        perf_counter=r_time,
    )

    return web.json_response(result)


async def _create_free_offer(
    profile: Profile,
    cred_def_id: str,
    connection_id: Optional[str] = None,
    auto_issue: bool = False,
    auto_remove: bool = False,
    preview_spec: Optional[dict] = None,
    comment: Optional[str] = None,
    trace_msg: Optional[bool] = None,
):
    """Create a credential offer and related exchange record."""

    credential_preview = CredentialPreview.deserialize(preview_spec)
    credential_proposal = CredentialProposal(
        comment=comment,
        credential_proposal=credential_preview,
        cred_def_id=cred_def_id,
    )
    credential_proposal.assign_trace_decorator(
        profile.settings,
        trace_msg,
    )
    credential_proposal_dict = credential_proposal.serialize()

    cred_ex_record = V10CredentialExchange(
        connection_id=connection_id,
        initiator=V10CredentialExchange.INITIATOR_SELF,
        role=V10CredentialExchange.ROLE_ISSUER,
        credential_definition_id=cred_def_id,
        credential_proposal_dict=credential_proposal_dict,
        auto_issue=auto_issue,
        auto_remove=auto_remove,
        trace=trace_msg,
    )

    credential_manager = CredentialManager(profile)

    (cred_ex_record, credential_offer_message) = await credential_manager.create_offer(
        cred_ex_record,
        counter_proposal=None,
        comment=comment,
    )

    return (cred_ex_record, credential_offer_message)


@docs(
    tags=["issue-credential v1.0"],
    summary="Create a credential offer, independent of any proposal or connection",
    deprecated=True,
)
@request_schema(V10CredentialConnFreeOfferRequestSchema())
@response_schema(V10CredentialExchangeSchema(), 200, description="")
@tenant_authentication
async def credential_exchange_create_free_offer(request: web.BaseRequest):
    """Request handler for creating free credential offer.

    Unlike with `send-offer`, this credential exchange is not tied to a specific
    connection. It must be dispatched out-of-band by the controller.

    Args:
        request: aiohttp request object

    Returns:
        The credential exchange record

    """
    r_time = get_timer()

    context: AdminRequestContext = request["context"]
    profile = context.profile
    body = await request.json()

    cred_def_id = body.get("cred_def_id")
    if not cred_def_id:
        raise web.HTTPBadRequest(reason="cred_def_id is required")

    auto_issue = body.get(
        "auto_issue", context.settings.get("debug.auto_respond_credential_request")
    )
    auto_remove = body.get(
        "auto_remove", not profile.settings.get("preserve_exchange_records")
    )
    comment = body.get("comment")
    preview_spec = body.get("credential_preview")
    if not preview_spec:
        raise web.HTTPBadRequest(reason="Missing credential_preview")

    trace_msg = body.get("trace")
    cred_ex_record = None
    try:
        (cred_ex_record, credential_offer_message) = await _create_free_offer(
            profile=profile,
            cred_def_id=cred_def_id,
            auto_issue=auto_issue,
            auto_remove=auto_remove,
            preview_spec=preview_spec,
            comment=comment,
            trace_msg=trace_msg,
        )
        result = cred_ex_record.serialize()
    except (
        BaseModelError,
        CredentialManagerError,
        IndyIssuerError,
        LedgerError,
        StorageError,
    ) as err:
        if cred_ex_record:
            async with profile.session() as session:
                await cred_ex_record.save_error_state(session, reason=err.roll_up)
        raise web.HTTPBadRequest(reason=err.roll_up)
    trace_event(
        context.settings,
        credential_offer_message,
        outcome="credential_exchange_create_free_offer.END",
        perf_counter=r_time,
    )
    return web.json_response(result)


@docs(
    tags=["issue-credential v1.0"],
    summary="Send holder a credential offer, independent of any proposal",
    deprecated=True,
)
@request_schema(V10CredentialFreeOfferRequestSchema())
@response_schema(V10CredentialExchangeSchema(), 200, description="")
@tenant_authentication
async def credential_exchange_send_free_offer(request: web.BaseRequest):
    """Request handler for sending free credential offer.

    An issuer initiates a such a credential offer, free from any
    holder-initiated corresponding credential proposal with preview.

    Args:
        request: aiohttp request object

    Returns:
        The credential exchange record

    """
    r_time = get_timer()

    context: AdminRequestContext = request["context"]
    profile = context.profile
    outbound_handler = request["outbound_message_router"]

    body = await request.json()

    connection_id = body.get("connection_id")
    cred_def_id = body.get("cred_def_id")
    if not cred_def_id:
        raise web.HTTPBadRequest(reason="cred_def_id is required")

    auto_issue = body.get(
        "auto_issue", context.settings.get("debug.auto_respond_credential_request")
    )
    auto_remove = body.get(
        "auto_remove", not profile.settings.get("preserve_exchange_records")
    )
    comment = body.get("comment")
    preview_spec = body.get("credential_preview")
    if not preview_spec:
        raise web.HTTPBadRequest(reason="Missing credential_preview")
    trace_msg = body.get("trace")

    cred_ex_record = None
    connection_record = None
    try:
        async with profile.session() as session:
            connection_record = await ConnRecord.retrieve_by_id(session, connection_id)
            if not connection_record.is_ready:
                raise web.HTTPForbidden(reason=f"Connection {connection_id} not ready")

        cred_ex_record, credential_offer_message = await _create_free_offer(
            profile=profile,
            cred_def_id=cred_def_id,
            connection_id=connection_id,
            auto_issue=auto_issue,
            auto_remove=auto_remove,
            preview_spec=preview_spec,
            comment=comment,
            trace_msg=trace_msg,
        )
        result = cred_ex_record.serialize()

    except (
        StorageNotFoundError,
        BaseModelError,
        CredentialManagerError,
        LedgerError,
    ) as err:
        if cred_ex_record:
            async with profile.session() as session:
                await cred_ex_record.save_error_state(session, reason=err.roll_up)
        # other party cannot yet receive a problem report about our failed protocol start
        raise web.HTTPBadRequest(reason=err.roll_up)

    await outbound_handler(credential_offer_message, connection_id=connection_id)

    trace_event(
        context.settings,
        credential_offer_message,
        outcome="credential_exchange_send_free_offer.END",
        perf_counter=r_time,
    )

    return web.json_response(result)


@docs(
    tags=["issue-credential v1.0"],
    summary="Send holder a credential offer in reference to a proposal with preview",
    deprecated=True,
)
@match_info_schema(CredExIdMatchInfoSchema())
@request_schema(V10CredentialBoundOfferRequestSchema())
@response_schema(V10CredentialExchangeSchema(), 200, description="")
@tenant_authentication
async def credential_exchange_send_bound_offer(request: web.BaseRequest):
    """Request handler for sending bound credential offer.

    A holder initiates this sequence with a credential proposal; this message
    responds with an offer bound to the proposal.

    Args:
        request: aiohttp request object

    Returns:
        The credential exchange record

    """
    r_time = get_timer()

    context: AdminRequestContext = request["context"]
    profile = context.profile
    outbound_handler = request["outbound_message_router"]

    body = await request.json() if request.body_exists else {}
    proposal_spec = body.get("counter_proposal")

    credential_exchange_id = request.match_info["cred_ex_id"]
    cred_ex_record = None
    connection_record = None
    try:
        async with profile.session() as session:
            try:
                cred_ex_record = await V10CredentialExchange.retrieve_by_id(
                    session, credential_exchange_id
                )
            except StorageNotFoundError as err:
                raise web.HTTPNotFound(reason=err.roll_up) from err

            if cred_ex_record.state != (
                V10CredentialExchange.STATE_PROPOSAL_RECEIVED
            ):  # check state here: manager call creates free offers too
                raise CredentialManagerError(
                    f"Credential exchange {cred_ex_record.credential_exchange_id} "
                    f"in {cred_ex_record.state} state "
                    f"(must be {V10CredentialExchange.STATE_PROPOSAL_RECEIVED})"
                )

            connection_id = cred_ex_record.connection_id
            connection_record = await ConnRecord.retrieve_by_id(session, connection_id)
            if not connection_record.is_ready:
                raise web.HTTPForbidden(reason=f"Connection {connection_id} not ready")

        credential_manager = CredentialManager(profile)
        (
            cred_ex_record,
            credential_offer_message,
        ) = await credential_manager.create_offer(
            cred_ex_record,
            counter_proposal=(
                CredentialProposal.deserialize(proposal_spec) if proposal_spec else None
            ),
            comment=None,
        )

        result = cred_ex_record.serialize()

    except (
        BaseModelError,
        CredentialManagerError,
        IndyIssuerError,
        LedgerError,
        StorageError,
    ) as err:
        if cred_ex_record:
            async with profile.session() as session:
                await cred_ex_record.save_error_state(session, reason=err.roll_up)
        await report_problem(
            err,
            ProblemReportReason.ISSUANCE_ABANDONED.value,
            web.HTTPBadRequest,
            cred_ex_record,
            outbound_handler,
        )

    await outbound_handler(credential_offer_message, connection_id=connection_id)

    trace_event(
        context.settings,
        credential_offer_message,
        outcome="credential_exchange_send_bound_offer.END",
        perf_counter=r_time,
    )

    return web.json_response(result)


@docs(
    tags=["issue-credential v1.0"],
    summary="Send issuer a credential request",
    deprecated=True,
)
@match_info_schema(CredExIdMatchInfoSchema())
@request_schema(V10CredentialExchangeAutoRemoveRequestSchema())
@response_schema(V10CredentialExchangeSchema(), 200, description="")
@tenant_authentication
async def credential_exchange_send_request(request: web.BaseRequest):
    """Request handler for sending credential request.

    Args:
        request: aiohttp request object

    Returns:
        The credential exchange record

    """
    r_time = get_timer()

    context: AdminRequestContext = request["context"]
    profile = context.profile
    outbound_handler = request["outbound_message_router"]

    credential_exchange_id = request.match_info["cred_ex_id"]

    try:
        body = await request.json() or {}
        auto_remove = body.get(
            "auto_remove", not profile.settings.get("preserve_exchange_records")
        )
    except JSONDecodeError:
        auto_remove = not profile.settings.get("preserve_exchange_records")

    cred_ex_record = None
    connection_record = None

    async with profile.session() as session:
        try:
            cred_ex_record = await V10CredentialExchange.retrieve_by_id(
                session, credential_exchange_id
            )
        except StorageNotFoundError as err:
            raise web.HTTPNotFound(reason=err.roll_up) from err

        # Fetch connection if exchange has record
        connection_record = None
        if cred_ex_record.connection_id:
            try:
                connection_record = await ConnRecord.retrieve_by_id(
                    session, cred_ex_record.connection_id
                )
            except StorageNotFoundError as err:
                raise web.HTTPBadRequest(reason=err.roll_up) from err

    if connection_record and not connection_record.is_ready:
        raise web.HTTPForbidden(
            reason=f"Connection {connection_record.connection_id} not ready"
        )

    if connection_record:
        holder_did = connection_record.my_did
    else:
        # Need to get the holder DID from the out of band record
        async with profile.session() as session:
            oob_record = await OobRecord.retrieve_by_tag_filter(
                session,
                {"invi_msg_id": cred_ex_record.credential_offer_dict._thread.pthid},
            )
            # Transform recipient key into did
            holder_did = default_did_from_verkey(oob_record.our_recipient_key)

    # assign the auto_remove flag from above...
    cred_ex_record.auto_remove = auto_remove

    try:
        credential_manager = CredentialManager(profile)
        (
            cred_ex_record,
            credential_request_message,
        ) = await credential_manager.create_request(cred_ex_record, holder_did)

        result = cred_ex_record.serialize()

    except (
        BaseModelError,
        CredentialManagerError,
        IndyHolderError,
        LedgerError,
        StorageError,
    ) as err:
        if cred_ex_record:
            async with profile.session() as session:
                await cred_ex_record.save_error_state(session, reason=err.roll_up)
        await report_problem(
            err,
            ProblemReportReason.ISSUANCE_ABANDONED.value,
            web.HTTPBadRequest,
            cred_ex_record,
            outbound_handler,
        )

    await outbound_handler(
        credential_request_message, connection_id=cred_ex_record.connection_id
    )

    trace_event(
        context.settings,
        credential_request_message,
        outcome="credential_exchange_send_request.END",
        perf_counter=r_time,
    )

    return web.json_response(result)


@docs(
    tags=["issue-credential v1.0"],
    summary="Send holder a credential",
    deprecated=True,
)
@match_info_schema(CredExIdMatchInfoSchema())
@request_schema(V10CredentialIssueRequestSchema())
@response_schema(V10CredentialExchangeSchema(), 200, description="")
@tenant_authentication
async def credential_exchange_issue(request: web.BaseRequest):
    """Request handler for sending credential.

    Args:
        request: aiohttp request object

    Returns:
        The credential exchange record

    """
    r_time = get_timer()

    context: AdminRequestContext = request["context"]
    profile = context.profile
    outbound_handler = request["outbound_message_router"]

    body = await request.json()
    comment = body.get("comment")

    credential_exchange_id = request.match_info["cred_ex_id"]

    cred_ex_record = None
    connection_record = None

    async with profile.session() as session:
        try:
            cred_ex_record = await V10CredentialExchange.retrieve_by_id(
                session, credential_exchange_id
            )
        except StorageNotFoundError as err:
            raise web.HTTPNotFound(reason=err.roll_up) from err

        # Fetch connection if exchange has record
        connection_record = None
        if cred_ex_record.connection_id:
            try:
                connection_record = await ConnRecord.retrieve_by_id(
                    session, cred_ex_record.connection_id
                )
            except StorageNotFoundError as err:
                raise web.HTTPBadRequest(reason=err.roll_up) from err

    if connection_record and not connection_record.is_ready:
        raise web.HTTPForbidden(
            reason=f"Connection {connection_record.connection_id} not ready"
        )

    try:
        credential_manager = CredentialManager(profile)
        (
            cred_ex_record,
            credential_issue_message,
        ) = await credential_manager.issue_credential(cred_ex_record, comment=comment)

        result = cred_ex_record.serialize()

    except (
        BaseModelError,
        CredentialManagerError,
        IndyIssuerError,
        LedgerError,
        StorageError,
    ) as err:
        if cred_ex_record:
            async with profile.session() as session:
                await cred_ex_record.save_error_state(session, reason=err.roll_up)
        await report_problem(
            err,
            ProblemReportReason.ISSUANCE_ABANDONED.value,
            web.HTTPBadRequest,
            cred_ex_record,
            outbound_handler,
        )

    await outbound_handler(
        credential_issue_message, connection_id=cred_ex_record.connection_id
    )

    trace_event(
        context.settings,
        credential_issue_message,
        outcome="credential_exchange_issue.END",
        perf_counter=r_time,
    )

    return web.json_response(result)


@docs(
    tags=["issue-credential v1.0"],
    summary="Store a received credential",
    deprecated=True,
)
@match_info_schema(CredExIdMatchInfoSchema())
@request_schema(V10CredentialStoreRequestSchema())
@response_schema(V10CredentialExchangeSchema(), 200, description="")
@tenant_authentication
async def credential_exchange_store(request: web.BaseRequest):
    """Request handler for storing credential.

    Args:
        request: aiohttp request object

    Returns:
        The credential exchange record

    """
    r_time = get_timer()

    context: AdminRequestContext = request["context"]
    profile = context.profile
    outbound_handler = request["outbound_message_router"]

    try:
        body = await request.json() or {}
        credential_id = body.get("credential_id")
    except JSONDecodeError:
        credential_id = None

    credential_exchange_id = request.match_info["cred_ex_id"]

    cred_ex_record = None
    connection_record = None

    async with profile.session() as session:
        try:
            cred_ex_record = await V10CredentialExchange.retrieve_by_id(
                session, credential_exchange_id
            )
        except StorageNotFoundError as err:
            raise web.HTTPNotFound(reason=err.roll_up) from err

        # Fetch connection if exchange has record
        if cred_ex_record.connection_id:
            try:
                connection_record = await ConnRecord.retrieve_by_id(
                    session, cred_ex_record.connection_id
                )
            except StorageNotFoundError as err:
                raise web.HTTPBadRequest(reason=err.roll_up) from err

    if connection_record and not connection_record.is_ready:
        raise web.HTTPForbidden(
            reason=f"Connection {connection_record.connection_id} not ready"
        )

    try:
        credential_manager = CredentialManager(profile)
        cred_ex_record = await credential_manager.store_credential(
            cred_ex_record,
            credential_id,
        )

    except (
        CredentialManagerError,
        IndyHolderError,
        StorageError,
    ) as err:  # treat failure to store as mangled on receipt hence protocol error
        if cred_ex_record:
            async with profile.session() as session:
                await cred_ex_record.save_error_state(session, reason=err.roll_up)
        await report_problem(
            err,
            ProblemReportReason.ISSUANCE_ABANDONED.value,
            web.HTTPBadRequest,
            cred_ex_record,
            outbound_handler,
        )

    try:  # protocol owes an ack
        (
            cred_ex_record,
            credential_ack_message,
        ) = await credential_manager.send_credential_ack(cred_ex_record)
        result = cred_ex_record.serialize()  # pick up state done

    except (
        BaseModelError,
        CredentialManagerError,
        StorageError,
    ) as err:
        # protocol finished OK: do not send problem report nor set record state error
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    trace_event(
        context.settings,
        credential_ack_message,
        outcome="credential_exchange_store.END",
        perf_counter=r_time,
    )

    return web.json_response(result)


@docs(
    tags=["issue-credential v1.0"],
    summary="Send a problem report for credential exchange",
    deprecated=True,
)
@match_info_schema(CredExIdMatchInfoSchema())
@request_schema(V10CredentialProblemReportRequestSchema())
@response_schema(IssueCredentialModuleResponseSchema(), 200, description="")
@tenant_authentication
async def credential_exchange_problem_report(request: web.BaseRequest):
    """Request handler for sending problem report.

    Args:
        request: aiohttp request object

    """
    context: AdminRequestContext = request["context"]
    outbound_handler = request["outbound_message_router"]

    credential_exchange_id = request.match_info["cred_ex_id"]
    body = await request.json()
    description = body["description"]

    try:
        async with context.profile.session() as session:
            cred_ex_record = await V10CredentialExchange.retrieve_by_id(
                session, credential_exchange_id
            )

            if not cred_ex_record.connection_id:
                raise web.HTTPBadRequest(
                    reason="No connection associated with credential exchange."
                )
            report = problem_report_for_record(cred_ex_record, description)
            await cred_ex_record.save_error_state(
                session,
                reason=f"created problem report: {description}",
            )
    except StorageNotFoundError as err:  # other party does not care about meta-problems
        raise web.HTTPNotFound(reason=err.roll_up) from err
    except StorageError as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    await outbound_handler(report, connection_id=cred_ex_record.connection_id)

    return web.json_response({})


@docs(
    tags=["issue-credential v1.0"],
    summary="Remove an existing credential exchange record",
    deprecated=True,
)
@match_info_schema(CredExIdMatchInfoSchema())
@response_schema(IssueCredentialModuleResponseSchema(), 200, description="")
@tenant_authentication
async def credential_exchange_remove(request: web.BaseRequest):
    """Request handler for removing a credential exchange record.

    Args:
        request: aiohttp request object

    """
    context: AdminRequestContext = request["context"]

    credential_exchange_id = request.match_info["cred_ex_id"]
    cred_ex_record = None
    try:
        async with context.profile.session() as session:
            cred_ex_record = await V10CredentialExchange.retrieve_by_id(
                session, credential_exchange_id
            )
            await cred_ex_record.delete_record(session)
    except StorageNotFoundError as err:  # not a protocol error
        raise web.HTTPNotFound(reason=err.roll_up) from err
    except StorageError as err:  # not a protocol error
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    return web.json_response({})


async def register(app: web.Application):
    """Register routes."""

    app.add_routes(
        [
            web.get(
                "/issue-credential/records", credential_exchange_list, allow_head=False
            ),
            web.post(
                "/issue-credential/create-offer", credential_exchange_create_free_offer
            ),
            web.get(
                "/issue-credential/records/{cred_ex_id}",
                credential_exchange_retrieve,
                allow_head=False,
            ),
            web.post("/issue-credential/create", credential_exchange_create),
            web.post("/issue-credential/send", credential_exchange_send),
            web.post(
                "/issue-credential/send-proposal", credential_exchange_send_proposal
            ),
            web.post("/issue-credential/send-offer", credential_exchange_send_free_offer),
            web.post(
                "/issue-credential/records/{cred_ex_id}/send-offer",
                credential_exchange_send_bound_offer,
            ),
            web.post(
                "/issue-credential/records/{cred_ex_id}/send-request",
                credential_exchange_send_request,
            ),
            web.post(
                "/issue-credential/records/{cred_ex_id}/issue",
                credential_exchange_issue,
            ),
            web.post(
                "/issue-credential/records/{cred_ex_id}/store",
                credential_exchange_store,
            ),
            web.post(
                "/issue-credential/records/{cred_ex_id}/problem-report",
                credential_exchange_problem_report,
            ),
            web.delete(
                "/issue-credential/records/{cred_ex_id}",
                credential_exchange_remove,
            ),
        ]
    )


def post_process_routes(app: web.Application):
    """Amend swagger API."""

    # Add top-level tags description
    if "tags" not in app._state["swagger_dict"]:
        app._state["swagger_dict"]["tags"] = []
    app._state["swagger_dict"]["tags"].append(
        {
            "name": "issue-credential v1.0",
            "description": "Credential issue v1.0",
            "externalDocs": {"description": "Specification", "url": SPEC_URI},
        }
    )



from unittest import IsolatedAsyncioTestCase

from ......messaging.request_context import RequestContext
from ......messaging.responder import MockResponder
from ......tests import mock
from ......transport.inbound.receipt import MessageReceipt
from ......utils.testing import create_test_profile
from ...messages.credential_proposal import CredentialProposal
from .. import credential_proposal_handler as test_module


class TestCredentialProposalHandler(IsolatedAsyncioTestCase):
    async def test_called(self):
        request_context = RequestContext.test_context(await create_test_profile())
        request_context.message_receipt = MessageReceipt()
        request_context.connection_record = mock.MagicMock()

        with mock.patch.object(
            test_module, "CredentialManager", autospec=True
        ) as mock_cred_mgr:
            mock_cred_mgr.return_value.receive_proposal = mock.CoroutineMock(
                return_value=mock.MagicMock()
            )
            mock_cred_mgr.return_value.receive_proposal.return_value.auto_offer = False
            request_context.message = CredentialProposal()
            request_context.connection_ready = True
            handler = test_module.CredentialProposalHandler()
            responder = MockResponder()
            await handler.handle(request_context, responder)

        mock_cred_mgr.assert_called_once_with(request_context.profile)
        mock_cred_mgr.return_value.receive_proposal.assert_called_once_with(
            request_context.message, request_context.connection_record.connection_id
        )
        assert not responder.messages

    async def test_called_auto_offer(self):
        request_context = RequestContext.test_context(await create_test_profile())
        request_context.message_receipt = MessageReceipt()
        request_context.connection_record = mock.MagicMock()

        with mock.patch.object(
            test_module, "CredentialManager", autospec=True
        ) as mock_cred_mgr:
            mock_cred_mgr.return_value.receive_proposal = mock.CoroutineMock(
                return_value=mock.MagicMock()
            )
            mock_cred_mgr.return_value.receive_proposal.return_value.auto_offer = True
            mock_cred_mgr.return_value.create_offer = mock.CoroutineMock(
                return_value=(None, "credential_offer_message")
            )
            request_context.message = CredentialProposal()
            request_context.connection_ready = True
            handler = test_module.CredentialProposalHandler()
            responder = MockResponder()
            await handler.handle(request_context, responder)

        mock_cred_mgr.assert_called_once_with(request_context.profile)
        mock_cred_mgr.return_value.receive_proposal.assert_called_once_with(
            request_context.message, request_context.connection_record.connection_id
        )
        messages = responder.messages
        assert len(messages) == 1
        (result, target) = messages[0]
        assert result == "credential_offer_message"
        assert target == {}

    async def test_called_auto_offer_x(self):
        request_context = RequestContext.test_context(await create_test_profile())
        request_context.message_receipt = MessageReceipt()
        request_context.connection_record = mock.MagicMock()

        with mock.patch.object(
            test_module, "CredentialManager", autospec=True
        ) as mock_cred_mgr:
            mock_cred_mgr.return_value.receive_proposal = mock.CoroutineMock(
                return_value=mock.MagicMock(save_error_state=mock.CoroutineMock())
            )
            mock_cred_mgr.return_value.receive_proposal.return_value.auto_offer = True
            mock_cred_mgr.return_value.create_offer = mock.CoroutineMock(
                side_effect=test_module.IndyIssuerError()
            )

            request_context.message = CredentialProposal()
            request_context.connection_ready = True
            handler = test_module.CredentialProposalHandler()
            responder = MockResponder()

            with (
                mock.patch.object(responder, "send_reply", mock.CoroutineMock()),
                mock.patch.object(
                    handler._logger, "exception", mock.MagicMock()
                ) as mock_log_exc,
            ):
                await handler.handle(request_context, responder)
                mock_log_exc.assert_called_once()

    async def test_called_not_ready(self):
        request_context = RequestContext.test_context(await create_test_profile())
        request_context.message_receipt = MessageReceipt()
        request_context.connection_record = mock.MagicMock()

        with mock.patch.object(
            test_module, "CredentialManager", autospec=True
        ) as mock_cred_mgr:
            mock_cred_mgr.return_value.receive_proposal = mock.CoroutineMock()
            request_context.message = CredentialProposal()
            request_context.connection_ready = False
            handler = test_module.CredentialProposalHandler()
            responder = MockResponder()
            with self.assertRaises(test_module.HandlerException) as err:
                await handler.handle(request_context, responder)
            assert (
                err.exception.message
                == "Connection used for credential proposal not ready"
            )

        assert not responder.messages

    async def test_called_no_connection(self):
        request_context = RequestContext.test_context(await create_test_profile())
        request_context.message_receipt = MessageReceipt()

        request_context.message = CredentialProposal()
        handler = test_module.CredentialProposalHandler()
        responder = MockResponder()
        with self.assertRaises(test_module.HandlerException) as err:
            await handler.handle(request_context, responder)
        assert (
            err.exception.message
            == "Connectionless not supported for credential proposal"
        )

        assert not responder.messages
